@@ -8,6 +8,9 @@
 
 #pragma once
 
+
+#define OSMO_CTRL_PORT_TRX	4236
+/* 4237 used by VTY interface */
 #define OSMO_CTRL_PORT_BTS	4238
 #define OSMO_CTRL_PORT_NITB_BSC	4249
 #define OSMO_CTRL_PORT_BSC_NAT	4250
@@ -17,4 +20,6 @@
 /* 4256 used by VTY interface */
 #define OSMO_CTRL_PORT_GGSN	4257
 #define OSMO_CTRL_PORT_HLR	4259
+#define OSMO_CTRL_PORT_HNBGW	4262
+#define OSMO_CTRL_PORT_GBPROXY	4263
 /* When adding/changing port numbers, keep docs and wiki in sync. See above. */
