@@ -1,9 +1,11 @@
 /* (C) 2009 by Harald Welte <laforge@gnumonks.org>
  * (C) 2012 Ivan Klyuchnikov
- * (C) 2015 by Sysmocom s.f.m.c. GmbH
+ * (C) 2015 by sysmocom - s.f.m.c. GmbH
  *
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -42,6 +44,7 @@
 
 #include <osmocom/core/bits.h>
 #include <osmocom/core/bitvec.h>
+#include <osmocom/core/panic.h>
 
 #define BITNUM_FROM_COMP(byte, bit)	((byte*8)+bit)
 
@@ -528,7 +531,7 @@ char bit_value_to_char(enum bit_value v)
 	case ONE: return '1';
 	case L: return 'L';
 	case H: return 'H';
-	default: abort();
+	default: osmo_panic("unexpected input in bit_value_to_char"); return 'X';
 	}
 }
 
