@@ -6,6 +6,8 @@
  *
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -53,7 +55,7 @@
  * implementing the GSM/GPRS/EGPRS channel coding (and decoding) as
  * specified in 3GPP TS 05.03 / 45.003.
  *
- * libosmocodec is developed as part of the Osmocom (Open Source Mobile
+ * libosmocoding is developed as part of the Osmocom (Open Source Mobile
  * Communications) project, a community-based, collaborative development
  * project to create Free and Open Source implementations of mobile
  * communications systems.  For more information about Osmocom, please
@@ -129,13 +131,6 @@
 #define EGPRS_DATA_C1		612
 #define EGPRS_DATA_C2		EGPRS_DATA_C1
 
-/* TS 101318 Chapter 5.1: 260 bits + 4bit sig */
-#define GSM_FR_BYTES	33
-/* TS 101318 Chapter 5.2: 112 bits, no sig */
-#define GSM_HR_BYTES	14
-/* TS 101318 Chapter 5.3: 244 bits + 4bit sig */
-#define GSM_EFR_BYTES	31
-
 /*! union across the three different EGPRS Uplink header types */
 union gprs_rlc_ul_hdr_egprs {
 	struct gprs_rlc_ul_header_egprs_1 type1;
@@ -533,16 +528,18 @@ const struct gsm0503_mcs_code gsm0503_mcs_dl_codes[EGPRS_NUM_MCS] = {
 	},
 };
 
-/*! Convolutional Decode + compute BER
+/*! Convolutional Decode + compute BER for punctured codes
  *  \param[in] code Description of Convolutional Code
  *  \param[in] input Input soft-bits (-127...127)
  *  \param[out] output bits
  *  \param[out] n_errors Number of bit-errors
  *  \param[out] n_bits_total Number of bits
+ *  \param[in] data_punc Puncturing mask array. Can be NULL.
  */
-static int osmo_conv_decode_ber(const struct osmo_conv_code *code,
+static int osmo_conv_decode_ber_punctured(const struct osmo_conv_code *code,
 	const sbit_t *input, ubit_t *output,
-	int *n_errors, int *n_bits_total)
+	int *n_errors, int *n_bits_total,
+	const uint8_t *data_punc)
 {
 	int res, i, coded_len;
 	ubit_t recoded[EGPRS_DATA_C_MAX];
@@ -558,7 +555,8 @@ static int osmo_conv_decode_ber(const struct osmo_conv_code *code,
 	if (n_errors) {
 		*n_errors = 0;
 		for (i = 0; i < coded_len; i++) {
-			if (!((recoded[i] && input[i] < 0) ||
+			if (((!data_punc) || (data_punc && !data_punc[i])) &&
+				!((recoded[i] && input[i] < 0) ||
 					(!recoded[i] && input[i] > 0)) )
 						*n_errors += 1;
 		}
@@ -570,6 +568,21 @@ static int osmo_conv_decode_ber(const struct osmo_conv_code *code,
 	return res;
 }
 
+/*! Convolutional Decode + compute BER for non-punctured codes
+ *  \param[in] code Description of Convolutional Code
+ *  \param[in] input Input soft-bits (-127...127)
+ *  \param[out] output bits
+ *  \param[out] n_errors Number of bit-errors
+ *  \param[out] n_bits_total Number of bits
+ */
+static int osmo_conv_decode_ber(const struct osmo_conv_code *code,
+	const sbit_t *input, ubit_t *output,
+	int *n_errors, int *n_bits_total)
+{
+	return osmo_conv_decode_ber_punctured(code, input, output,
+		n_errors, n_bits_total, NULL);
+}
+
 /*! convenience wrapper for decoding coded bits
  *  \param[out] l2_data caller-allocated buffer for L2 Frame
  *  \param[in] cB 456 coded (soft) bits as per TS 05.03 4.1.3
@@ -889,7 +902,8 @@ static int egprs_decode_data(uint8_t *l2_data, const sbit_t *c,
 			C[i] = 0;
 	}
 
-	osmo_conv_decode_ber(code->data_conv, C, u, n_errors, n_bits_total);
+	osmo_conv_decode_ber_punctured(code->data_conv, C, u,
+		n_errors, n_bits_total, code->data_punc[p]);
 	rc = osmo_crc16gen_check_bits(&gsm0503_mcs_crc12, u,
 		data_len, u + data_len);
 	if (rc)
@@ -913,7 +927,7 @@ static int egprs_decode_data(uint8_t *l2_data, const sbit_t *c,
  *  \param[in] nbits number of bits in \a bursts
  *  \param usf_p unused argument ?!?
  *  \param[out] n_errors number of detected bit-errors
- *  \param[out] n_bits_total total number of dcoded bits
+ *  \param[out] n_bits_total total number of decoded bits
  *  \returns 0 on success; negative on error */
 int gsm0503_pdtch_egprs_decode(uint8_t *l2_data, const sbit_t *bursts, uint16_t nbits,
 	uint8_t *usf_p, int *n_errors, int *n_bits_total)
@@ -1743,16 +1757,16 @@ static void tch_efr_unreorder(ubit_t *s, ubit_t *p, const ubit_t *w)
 
 	memcpy(s, w, 71);
 	sum = s[69] + w[71] + w[72];
-	s[69] = (sum > 2);
+	s[69] = (sum >= 2);
 	memcpy(s + 71, w + 73, 50);
 	sum = s[119] + w[123] + w[124];
-	s[119] = (sum > 2);
+	s[119] = (sum >= 2);
 	memcpy(s + 121, w + 125, 53);
 	sum = s[172] + w[178] + w[179];
 	s[172] = (sum > 2);
 	memcpy(s + 174, w + 180, 50);
-	sum = s[220] + w[230] + w[231];
-	s[222] = (sum > 2);
+	sum = s[222] + w[230] + w[231];
+	s[222] = (sum >= 2);
 	memcpy(s + 224, w + 232, 20);
 	memcpy(p, w + 252, 8);
 }
@@ -1778,7 +1792,7 @@ static void tch_amr_unmerge(ubit_t *d, ubit_t *p, const ubit_t *u, int len, int
  *  \param[in] efr Is this channel using EFR (1) or FR (0)
  *  \param[out] n_errors Number of detected bit errors
  *  \param[out] n_bits_total Total number of bits
- *  \returns length of bytes used in \a tch_data output buffer */
+ *  \returns length of bytes used in \a tch_data output buffer; negative on error */
 int gsm0503_tch_fr_decode(uint8_t *tch_data, const sbit_t *bursts,
 	int net_order, int efr, int *n_errors, int *n_bits_total)
 {
@@ -1927,7 +1941,7 @@ coding_efr_fr:
  *  \param[in] odd Odd (1) or even (0) frame number
  *  \param[out] n_errors Number of detected bit errors
  *  \param[out] n_bits_total Total number of bits
- *  \returns length of bytes used in \a tch_data output buffer */
+ *  \returns length of bytes used in \a tch_data output buffer; negative on error */
 int gsm0503_tch_hr_decode(uint8_t *tch_data, const sbit_t *bursts, int odd,
 	int *n_errors, int *n_bits_total)
 {
@@ -2068,7 +2082,9 @@ int gsm0503_tch_hr_encode(ubit_t *bursts, const uint8_t *tch_data, int len)
  *  \param[out] cmr Output in \a codec_mode_req = 1
  *  \param[out] n_errors Number of detected bit errors
  *  \param[out] n_bits_total Total number of bits
- *  \returns length of bytes used in \a tch_data output buffer */
+ *  \returns (>=4) length of bytes used in \a tch_data output buffer; ([0,3])
+ *  	     codec out of range; negative on error
+ */
 int gsm0503_tch_afs_decode(uint8_t *tch_data, const sbit_t *bursts,
 	int codec_mode_req, uint8_t *codec, int codecs, uint8_t *ft,
 	uint8_t *cmr, int *n_errors, int *n_bits_total)
@@ -2445,7 +2461,9 @@ invalid_length:
  *  \param[out] cmr Output in \a codec_mode_req = 1
  *  \param[out] n_errors Number of detected bit errors
  *  \param[out] n_bits_total Total number of bits
- *  \returns length of bytes used in \a tch_data output buffer */
+ *  \returns (>=4) length of bytes used in \a tch_data output buffer; ([0,3])
+ *  	     codec out of range; negative on error
+ */
 int gsm0503_tch_ahs_decode(uint8_t *tch_data, const sbit_t *bursts, int odd,
 	int codec_mode_req, uint8_t *codec, int codecs, uint8_t *ft,
 	uint8_t *cmr, int *n_errors, int *n_bits_total)
@@ -2822,13 +2840,49 @@ invalid_length:
  * b(0) = MSB of PLMN colour code
  * b(5) = LSB of BS colour code
  */
-static int rach_apply_bsic(ubit_t *d, uint8_t bsic)
+static inline void rach_apply_bsic(ubit_t *d, uint8_t bsic, uint8_t start)
 {
 	int i;
 
 	/* Apply it */
 	for (i = 0; i < 6; i++)
-		d[8 + i] ^= ((bsic >> (5 - i)) & 1);
+		d[start + i] ^= ((bsic >> (5 - i)) & 1);
+}
+
+static inline int16_t rach_decode_ber(const sbit_t *burst, uint8_t bsic, bool is_11bit,
+				      int *n_errors, int *n_bits_total)
+{
+	ubit_t conv[17];
+	uint8_t ra[2] = { 0 }, nbits = is_11bit ? 11 : 8;
+	int rv;
+
+	osmo_conv_decode_ber(is_11bit ? &gsm0503_rach_ext : &gsm0503_rach, burst, conv,
+			     n_errors, n_bits_total);
+
+	rach_apply_bsic(conv, bsic, nbits);
+
+	rv = osmo_crc8gen_check_bits(&gsm0503_rach_crc6, conv, nbits, conv + nbits);
+	if (rv)
+		return -1;
+
+	osmo_ubit2pbit_ext(ra, 0, conv, 0, nbits, 1);
+
+	return is_11bit ? osmo_load16le(ra) : ra[0];
+}
+
+/*! Decode the Extended (11-bit) RACH according to 3GPP TS 45.003
+ *  \param[out] ra output buffer for RACH data
+ *  \param[in] burst Input burst data
+ *  \param[in] bsic BSIC used in this cell
+ *  \returns 0 on success; negative on error (e.g. CRC error) */
+int gsm0503_rach_ext_decode(uint16_t *ra, const sbit_t *burst, uint8_t bsic)
+{
+	int16_t r = rach_decode_ber(burst, bsic, true, NULL, NULL);
+
+	if (r < 0)
+		return r;
+
+	*ra = r;
 
 	return 0;
 }
@@ -2840,18 +2894,48 @@ static int rach_apply_bsic(ubit_t *d, uint8_t bsic)
  *  \returns 0 on success; negative on error (e.g. CRC error) */
 int gsm0503_rach_decode(uint8_t *ra, const sbit_t *burst, uint8_t bsic)
 {
-	ubit_t conv[14];
-	int rv;
+	int16_t r = rach_decode_ber(burst, bsic, false, NULL, NULL);
+	if (r < 0)
+		return r;
 
-	osmo_conv_decode(&gsm0503_rach, burst, conv);
+	*ra = r;
+	return 0;
+}
 
-	rach_apply_bsic(conv, bsic);
+/*! Decode the Extended (11-bit) RACH according to 3GPP TS 45.003
+ *  \param[out] ra output buffer for RACH data
+ *  \param[in] burst Input burst data
+ *  \param[in] bsic BSIC used in this cell
+ *  \param[out] n_errors Number of detected bit errors
+ *  \param[out] n_bits_total Total number of bits
+ *  \returns 0 on success; negative on error (e.g. CRC error) */
+int gsm0503_rach_ext_decode_ber(uint16_t *ra, const sbit_t *burst, uint8_t bsic,
+				int *n_errors, int *n_bits_total)
+{
+	int16_t r = rach_decode_ber(burst, bsic, true, n_errors, n_bits_total);
+	if (r < 0)
+		return r;
 
-	rv = osmo_crc8gen_check_bits(&gsm0503_rach_crc6, conv, 8, conv + 8);
-	if (rv)
-		return -1;
+	*ra = r;
+	return 0;
+}
 
-	osmo_ubit2pbit_ext(ra, 0, conv, 0, 8, 1);
+/*! Decode the (8-bit) RACH according to TS 05.03
+ *  \param[out] ra output buffer for RACH data
+ *  \param[in] burst Input burst data
+ *  \param[in] bsic BSIC used in this cell
+ *  \param[out] n_errors Number of detected bit errors
+ *  \param[out] n_bits_total Total number of bits
+ *  \returns 0 on success; negative on error (e.g. CRC error) */
+int gsm0503_rach_decode_ber(uint8_t *ra, const sbit_t *burst, uint8_t bsic,
+			    int *n_errors, int *n_bits_total)
+{
+	int16_t r = rach_decode_ber(burst, bsic, false, n_errors, n_bits_total);
+
+	if (r < 0)
+		return r;
+
+	*ra = r;
 
 	return 0;
 }
@@ -2863,15 +2947,33 @@ int gsm0503_rach_decode(uint8_t *ra, const sbit_t *burst, uint8_t bsic)
  *  \returns 0 on success; negative on error */
 int gsm0503_rach_encode(ubit_t *burst, const uint8_t *ra, uint8_t bsic)
 {
-	ubit_t conv[14];
+	return gsm0503_rach_ext_encode(burst, *ra, bsic, false);
+}
 
-	osmo_pbit2ubit_ext(conv, 0, ra, 0, 8, 1);
+/*! Encode the Extended (11-bit) or regular (8-bit) RACH according to 3GPP TS 45.003
+ *  \param[out] burst Caller-allocated output burst buffer
+ *  \param[in] ra11 Input RACH data
+ *  \param[in] bsic BSIC used in this cell
+ *  \param[in] is_11bit whether given RA is 11 bit or not
+ *  \returns 0 on success; negative on error */
+int gsm0503_rach_ext_encode(ubit_t *burst, uint16_t ra11, uint8_t bsic, bool is_11bit)
+{
+	ubit_t conv[17];
+	uint8_t ra[2] = { 0 }, nbits = 8;
 
-	osmo_crc8gen_set_bits(&gsm0503_rach_crc6, conv, 8, conv + 8);
+	if (is_11bit) {
+		osmo_store16le(ra11, ra);
+		nbits = 11;
+	} else
+		ra[0] = (uint8_t)ra11;
 
-	rach_apply_bsic(conv, bsic);
+	osmo_pbit2ubit_ext(conv, 0, ra, 0, nbits, 1);
 
-	osmo_conv_encode(&gsm0503_rach, conv, burst);
+	osmo_crc8gen_set_bits(&gsm0503_rach_crc6, conv, nbits, conv + nbits);
+
+	rach_apply_bsic(conv, bsic, nbits);
+
+	osmo_conv_encode(is_11bit ? &gsm0503_rach_ext : &gsm0503_rach, conv, burst);
 
 	return 0;
 }
