@@ -3,9 +3,13 @@
 /*
  * (C) 2010-2011 by Daniel Willmann <daniel@totalueberwachung.de>
  * (C) 2010-2011 by On-Waves
+ * (C) 2014 by Harald Welte <laforge@gnumonks.org>
+ * (C) 2016-2017 by sysmocom - s.f.m.c. GmbH
  *
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -71,6 +75,11 @@ struct lookup_helper {
 };
 static LLIST_HEAD(ctrl_lookup_helpers);
 
+/*! Parse ascii-encoded decimal number at vline[i]
+ *  \param[in] vline vector containing a tokenized line
+ *  \param[in] i index into the vector \a vline
+ *  \param[out] num parsed decimal integer number at vline[i]
+ *  \returns 1 on success; 0 in case of error */
 int ctrl_parse_get_num(vector vline, int i, long *num)
 {
 	char *token, *tmp;
@@ -90,7 +99,10 @@ int ctrl_parse_get_num(vector vline, int i, long *num)
 	return 1;
 }
 
-/* Send command to all  */
+/*! Send a CTRL command to all connections.
+ *  \param[in] ctrl global control handle
+ *  \param[in] cmd command to send to all connections in \ctrl
+ *  \returns number of times the command has been sent */
 int ctrl_cmd_send_to_all(struct ctrl_handle *ctrl, struct ctrl_cmd *cmd)
 {
 	struct ctrl_connection *ccon;
@@ -105,6 +117,10 @@ int ctrl_cmd_send_to_all(struct ctrl_handle *ctrl, struct ctrl_cmd *cmd)
 	return ret;
 }
 
+/*! Encode a CTRL command and append it to the given write queue
+ *  \param[inout] queue write queue to which encoded \a cmd shall be appended
+ *  \param[in] cmd decoded command representation
+ *  \returns 0 in case of success; negative on error */
 int ctrl_cmd_send(struct osmo_wqueue *queue, struct ctrl_cmd *cmd)
 {
 	int ret;
@@ -148,6 +164,9 @@ int ctrl_cmd_send_trap(struct ctrl_handle *ctrl, const char *name, char *value)
 	return r;
 }
 
+/*! Copy given \a cmd and convert copy to CTRL_TYPE_TRAP.
+ *  \param[in] cmd command to be copied
+ *  \returns pointer to newly-allocated copy of type TRAP. Allocated as sibling of \a cmd */
 struct ctrl_cmd *ctrl_cmd_trap(struct ctrl_cmd *cmd)
 {
 	struct ctrl_cmd *trap;
@@ -164,6 +183,10 @@ struct ctrl_cmd *ctrl_cmd_trap(struct ctrl_cmd *cmd)
 static void control_close_conn(struct ctrl_connection *ccon)
 {
 	struct ctrl_cmd_def *cd, *cd2;
+	char *name = osmo_sock_get_name(ccon, ccon->write_queue.bfd.fd);
+
+	LOGP(DLCTRL, LOGL_INFO, "close()d CTRL connection %s\n", name);
+	talloc_free(name);
 
 	osmo_wqueue_clear(&ccon->write_queue);
 	close(ccon->write_queue.bfd.fd);
@@ -222,7 +245,7 @@ int ctrl_cmd_handle(struct ctrl_handle *ctrl, struct ctrl_cmd *cmd,
 	if (!request)
 		goto err;
 
-	for (i=0;i<strlen(request);i++) {
+	for (i = 0; i < strlen(request); i++) {
 		if (request[i] == '.')
 			request[i] = ' ';
 	}
@@ -234,7 +257,7 @@ int ctrl_cmd_handle(struct ctrl_handle *ctrl, struct ctrl_cmd *cmd,
 		goto err;
 	}
 
-	for (i=0;i<vector_active(vline);i++) {
+	for (i = 0; i < vector_active(vline); i++) {
 		struct lookup_helper *lh;
 		int rc;
 
@@ -266,7 +289,7 @@ int ctrl_cmd_handle(struct ctrl_handle *ctrl, struct ctrl_cmd *cmd,
 			return ret;
 		default: /* If we're here the rest must be the command */
 			cmdvec = vector_init(vector_active(vline)-i);
-			for (j=i; j<vector_active(vline); j++) {
+			for (j = i; j < vector_active(vline); j++) {
 				vector_set(cmdvec, vector_slot(vline, j));
 			}
 
@@ -319,87 +342,129 @@ err:
 
 static int handle_control_read(struct osmo_fd * bfd)
 {
-	int ret = -1;
+	int ret;
 	struct osmo_wqueue *queue;
 	struct ctrl_connection *ccon;
-	struct ipaccess_head *iph;
-	struct ipaccess_head_ext *iph_ext;
 	struct msgb *msg = NULL;
-	struct ctrl_cmd *cmd;
 	struct ctrl_handle *ctrl = bfd->data;
 
 	queue = container_of(bfd, struct osmo_wqueue, bfd);
 	ccon = container_of(queue, struct ctrl_connection, write_queue);
 
 	ret = ipa_msg_recv_buffered(bfd->fd, &msg, &ccon->pending_msg);
-	if (ret <= 0) {
-		if (ret == -EAGAIN)
-			return 0;
-		if (ret == 0)
-			LOGP(DLCTRL, LOGL_INFO, "The control connection was closed\n");
-		else
-			LOGP(DLCTRL, LOGL_ERROR, "Failed to parse ip access message: %d\n", ret);
-
-		goto err;
+	if (ret == 0) {
+		/* msg was already discarded. */
+		goto close_fd;
+	} else if (ret == -EAGAIN) {
+		/* received part of a message, it is stored in ccon->pending_msg and there's
+		 * nothing left to do now. */
+		return 0;
+	} else if (ret < 0) {
+		LOGP(DLCTRL, LOGL_ERROR, "Failed to parse ip access message: %d\n", ret);
+		return 0;
 	}
 
+	ret = ctrl_handle_msg(ctrl, ccon, msg);
+	msgb_free(msg);
+	if (ret)
+		goto close_fd;
+
+	return 0;
+
+close_fd:
+	control_close_conn(ccon);
+	return -EBADF;
+}
+
+/*! Handle a received CTRL command contained in a \ref msgb.
+ *  \param[in] ctrl CTRL interface handle
+ *  \param[in] ccon CTRL connection through which the command was received
+ *  \param[in] msg message buffer containing CTRL command including IPA+IPA_EXT headers
+ *  \returns 0 on success; negative on error */
+int ctrl_handle_msg(struct ctrl_handle *ctrl, struct ctrl_connection *ccon, struct msgb *msg)
+{
+	struct ctrl_cmd *cmd;
+	bool parse_failed;
+	struct ipaccess_head *iph;
+	struct ipaccess_head_ext *iph_ext;
+	int result;
+
 	if (msg->len < sizeof(*iph) + sizeof(*iph_ext)) {
 		LOGP(DLCTRL, LOGL_ERROR, "The message is too short.\n");
-		goto err;
+		return -EINVAL;
 	}
 
 	iph = (struct ipaccess_head *) msg->data;
 	if (iph->proto != IPAC_PROTO_OSMO) {
 		LOGP(DLCTRL, LOGL_ERROR, "Protocol mismatch. We got 0x%x\n", iph->proto);
-		goto err;
+		return -EINVAL;
 	}
 
 	iph_ext = (struct ipaccess_head_ext *) iph->data;
 	if (iph_ext->proto != IPAC_PROTO_EXT_CTRL) {
 		LOGP(DLCTRL, LOGL_ERROR, "Extended protocol mismatch. We got 0x%x\n", iph_ext->proto);
-		goto err;
+		return -EINVAL;
 	}
 
 	msg->l2h = iph_ext->data;
 
-	cmd = ctrl_cmd_parse(ccon, msg);
+	cmd = ctrl_cmd_parse3(ccon, msg, &parse_failed);
 
-	if (cmd) {
-		cmd->ccon = ccon;
-		if (ctrl_cmd_handle(ctrl, cmd, ctrl->data) != CTRL_CMD_HANDLED) {
-			ctrl_cmd_send(queue, cmd);
-			talloc_free(cmd);
-		}
-	} else {
+	if (!cmd) {
+		/* should never happen */
 		cmd = talloc_zero(ccon, struct ctrl_cmd);
 		if (!cmd)
-			goto err;
+			return -ENOMEM;
 		LOGP(DLCTRL, LOGL_ERROR, "Command parser error.\n");
 		cmd->type = CTRL_TYPE_ERROR;
 		cmd->id = "err";
 		cmd->reply = "Command parser error.";
-		ctrl_cmd_send(queue, cmd);
-		talloc_free(cmd);
 	}
 
-	msgb_free(msg);
-	return 0;
+	/* In case of error, reply with the error message right away. */
+	if (cmd->type == CTRL_TYPE_ERROR && parse_failed)
+		goto send_reply;
 
-err:
-	control_close_conn(ccon);
-	msgb_free(msg);
-	return ret;
+	cmd->ccon = ccon;
+	result = ctrl_cmd_handle(ctrl, cmd, ctrl->data);
+
+
+	if (cmd->defer) {
+		/* The command is still stored as ctrl_cmd_def.cmd, in the def_cmds list.
+		 * Just leave hanging for deferred handling. Reply will happen later. */
+		return 0;
+	}
+
+	/* On CTRL_CMD_HANDLED, no reply needs to be sent back. */
+	if (result == CTRL_CMD_HANDLED)
+		goto just_free;
+
+send_reply:
+	/* There is a reply or error that should be reported back to the sender. */
+	ctrl_cmd_send(&ccon->write_queue, cmd);
+just_free:
+	talloc_free(cmd);
+	return 0;
 }
 
 static int control_write_cb(struct osmo_fd *bfd, struct msgb *msg)
 {
+	struct osmo_wqueue *queue;
+	struct ctrl_connection *ccon;
 	int rc;
 
-	rc = write(bfd->fd, msg->data, msg->len);
-	if (rc != msg->len)
-		LOGP(DLCTRL, LOGL_ERROR, "Failed to write message to the control connection.\n");
+	queue = container_of(bfd, struct osmo_wqueue, bfd);
+	ccon = container_of(queue, struct ctrl_connection, write_queue);
 
-	return rc;
+	rc = write(bfd->fd, msg->data, msg->len);
+	if (rc == 0) {
+		control_close_conn(ccon);
+		return -EBADF;
+	}
+	if (rc != msg->len)
+		LOGP(DLCTRL, LOGL_ERROR, "Failed to write message to the CTRL connection.\n");
+
+	return 0;
 }
 
 /*! Allocate CTRL connection
@@ -432,20 +497,17 @@ static int listen_fd_cb(struct osmo_fd *listen_bfd, unsigned int what)
 	int ret, fd, on;
 	struct ctrl_handle *ctrl;
 	struct ctrl_connection *ccon;
-	struct sockaddr_in sa;
-	socklen_t sa_len = sizeof(sa);
+	char *name;
 
 
 	if (!(what & BSC_FD_READ))
 		return 0;
 
-	fd = accept(listen_bfd->fd, (struct sockaddr *) &sa, &sa_len);
+	fd = accept(listen_bfd->fd, NULL, NULL);
 	if (fd < 0) {
 		perror("accept");
 		return fd;
 	}
-	LOGP(DLCTRL, LOGL_INFO, "accept()ed new control connection from %s\n",
-		inet_ntoa(sa.sin_addr));
 
 #ifdef TCP_NODELAY
 	on = 1;
@@ -464,6 +526,9 @@ static int listen_fd_cb(struct osmo_fd *listen_bfd, unsigned int what)
 		return -1;
 	}
 
+	name = osmo_sock_get_name(ccon, fd);
+	LOGP(DLCTRL, LOGL_INFO, "accept()ed new CTRL connection from %s\n", name);
+
 	ccon->write_queue.bfd.fd = fd;
 	ccon->write_queue.bfd.when = BSC_FD_READ;
 
@@ -479,10 +544,13 @@ static int listen_fd_cb(struct osmo_fd *listen_bfd, unsigned int what)
 	return ret;
 }
 
-static uint64_t get_rate_ctr_value(const struct rate_ctr *ctr, int intv)
+static uint64_t get_rate_ctr_value(const struct rate_ctr *ctr, int intv, const char *grp)
 {
-	if (intv >= RATE_CTR_INTV_NUM)
+	if (intv >= RATE_CTR_INTV_NUM) {
+		LOGP(DLCTRL, LOGL_ERROR, "Unexpected interval value %d while trying to get rate counter value in %s\n",
+		     intv, grp);
 		return 0;
+	}
 
 	/* Absolute value */
 	if (intv == -1) {
@@ -492,42 +560,32 @@ static uint64_t get_rate_ctr_value(const struct rate_ctr *ctr, int intv)
 	}
 }
 
-static char *get_all_rate_ctr_in_group(void *ctx, const struct rate_ctr_group *ctrg, int intv)
-{
-	int i;
-	char *counters = talloc_strdup(ctx, "");
-	if (!counters)
-		return NULL;
-
-	for (i=0;i<ctrg->desc->num_ctr;i++) {
-		counters = talloc_asprintf_append(counters, "\n%s.%u.%s %"PRIu64,
-			ctrg->desc->group_name_prefix, ctrg->idx,
-			ctrg->desc->ctr_desc[i].name,
-			get_rate_ctr_value(&ctrg->ctr[i], intv));
-		if (!counters)
-			return NULL;
-	}
-	return counters;
-}
-
 static int get_rate_ctr_group_idx(const struct rate_ctr_group *ctrg, int intv, struct ctrl_cmd *cmd)
 {
-	char *counters;
-
-	counters = get_all_rate_ctr_in_group(cmd, ctrg, intv);
-	if (!counters)
-		goto oom;
-
-	cmd->reply = talloc_asprintf(cmd, "All counters in %s.%u%s",
-			ctrg->desc->group_name_prefix, ctrg->idx, counters);
-	talloc_free(counters);
-	if (!cmd->reply)
-		goto oom;
+	unsigned int i;
+	for (i = 0; i < ctrg->desc->num_ctr; i++) {
+		ctrl_cmd_reply_printf(cmd, "%s %"PRIu64";", ctrg->desc->ctr_desc[i].name,
+				      get_rate_ctr_value(&ctrg->ctr[i], intv, ctrg->desc->group_name_prefix));
+		if (!cmd->reply) {
+			cmd->reply = "OOM";
+			return CTRL_CMD_ERROR;
+		}
+	}
 
 	return CTRL_CMD_REPLY;
-oom:
-	cmd->reply = "OOM.";
-	return CTRL_CMD_ERROR;
+}
+
+static int ctrl_rate_ctr_group_handler(struct rate_ctr_group *ctrg, void *data)
+{
+	struct ctrl_cmd *cmd = data;
+
+	cmd->reply = talloc_asprintf_append(cmd->reply, "%s.%u;", ctrg->desc->group_name_prefix, ctrg->idx);
+	if (!cmd->reply) {
+		cmd->reply = "OOM";
+		return -1;
+	}
+
+	return 0;
 }
 
 /* rate_ctr */
@@ -570,9 +628,14 @@ static int get_rate_ctr(struct ctrl_cmd *cmd, void *data)
 		intv = RATE_CTR_INTV_HOUR;
 	} else if (!strcmp(interval, "per_day")) {
 		intv = RATE_CTR_INTV_DAY;
+	} else if (!strcmp(interval, "*")) {
+		intv = rate_ctr_for_each_group(ctrl_rate_ctr_group_handler, cmd);
+		if (intv < 0)
+			return CTRL_CMD_ERROR;
+		return CTRL_CMD_REPLY;
 	} else {
 		talloc_free(dup);
-		cmd->reply = "Wrong interval.";
+		cmd->reply = "Wrong interval. Expecting 'per_sec', 'per_min', 'per_hour', 'per_day' or 'abs' value.";
 		goto err;
 	}
 
@@ -608,7 +671,7 @@ static int get_rate_ctr(struct ctrl_cmd *cmd, void *data)
 
 	talloc_free(dup);
 
-	cmd->reply = talloc_asprintf(cmd, "%"PRIu64, get_rate_ctr_value(ctr, intv));
+	cmd->reply = talloc_asprintf(cmd, "%"PRIu64, get_rate_ctr_value(ctr, intv, ctrg->desc->group_name_prefix));
 	if (!cmd->reply)
 		goto oom;
 
@@ -884,13 +947,16 @@ struct ctrl_cmd *ctrl_cmd_exec_from_string(struct ctrl_handle *ch, const char *c
 	osmo_strlcpy((char *)msg->data, cmdstr, msgb_tailroom(msg));
 	msgb_put(msg, strlen(cmdstr));
 
-	cmd = ctrl_cmd_parse(ch, msg);
+	cmd = ctrl_cmd_parse2(ch, msg);
 	msgb_free(msg);
 	if (!cmd)
 		return NULL;
-	if (ctrl_cmd_handle(ch, cmd, NULL) < 0) {
+	if (cmd->type == CTRL_TYPE_ERROR)
+		return cmd;
+	if (ctrl_cmd_handle(ch, cmd, NULL) == CTRL_CMD_HANDLED) {
+		/* No reply should be sent back. */
 		talloc_free(cmd);
-		return NULL;
+		cmd = NULL;
 	}
 	return cmd;
 }
