@@ -3,6 +3,8 @@
 /*
  * (C) 2016 by Harald Welte <laforge@gnumonks.org>
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2 of the License, or
@@ -117,7 +119,12 @@ struct osmo_fsm_inst *osmo_fsm_inst_find_by_name(const struct osmo_fsm *fsm,
 {
 	struct osmo_fsm_inst *fi;
 
+	if (!name)
+		return NULL;
+
 	llist_for_each_entry(fi, &fsm->instances, list) {
+		if (!fi->name)
+			continue;
 		if (!strcmp(name, fi->name))
 			return fi;
 	}
@@ -152,6 +159,8 @@ int osmo_fsm_register(struct osmo_fsm *fsm)
 	}
 	if (osmo_fsm_find_by_name(fsm->name))
 		return -EEXIST;
+	if (fsm->event_names == NULL)
+		LOGP(DLGLOBAL, LOGL_ERROR, "FSM '%s' has no event names! Please fix!\n", fsm->name);
 	llist_add_tail(&fsm->list, &osmo_g_fsms);
 	INIT_LLIST_HEAD(&fsm->instances);
 
@@ -181,24 +190,91 @@ static void fsm_tmr_cb(void *data)
 
 	if (fsm->timer_cb) {
 		int rc = fsm->timer_cb(fi);
-		if (rc != 1) {
-			fi->T = 0;
+		if (rc != 1)
+			/* We don't actually know whether fi exists anymore.
+			 * Make sure to not access it and return right away. */
 			return;
-		}
+		/* The timer_cb told us to terminate, so we can safely assume
+		 * that fi still exists. */
 		LOGPFSM(fi, "timer_cb requested termination\n");
 	} else
 		LOGPFSM(fi, "No timer_cb, automatic termination\n");
 
 	/* if timer_cb returns 1 or there is no timer_cb */
-	fi->T = 0;
 	osmo_fsm_inst_term(fi, OSMO_FSM_TERM_TIMEOUT, &T);
 }
 
+/*! Change id of the FSM instance
+ * \param[in] fi FSM instance
+ * \param[in] id new ID
+ * \returns 0 if the ID was updated, otherwise -EINVAL
+ */
+int osmo_fsm_inst_update_id(struct osmo_fsm_inst *fi, const char *id)
+{
+	if (!id)
+		return osmo_fsm_inst_update_id_f(fi, NULL);
+	else
+		return osmo_fsm_inst_update_id_f(fi, "%s", id);
+}
+
+static void update_name(struct osmo_fsm_inst *fi)
+{
+	if (fi->name)
+		talloc_free((char*)fi->name);
+
+	if (!fsm_log_addr) {
+		if (fi->id)
+			fi->name = talloc_asprintf(fi, "%s(%s)", fi->fsm->name, fi->id);
+		else
+			fi->name = talloc_asprintf(fi, "%s", fi->fsm->name);
+	} else {
+		if (fi->id)
+			fi->name = talloc_asprintf(fi, "%s(%s)[%p]", fi->fsm->name, fi->id, fi);
+		else
+			fi->name = talloc_asprintf(fi, "%s[%p]", fi->fsm->name, fi);
+	}
+}
+
+/*! Change id of the FSM instance using a string format.
+ * \param[in] fi FSM instance.
+ * \param[in] fmt format string to compose new ID.
+ * \param[in] ... variable argument list for format string.
+ * \returns 0 if the ID was updated, otherwise -EINVAL.
+ */
+int osmo_fsm_inst_update_id_f(struct osmo_fsm_inst *fi, const char *fmt, ...)
+{
+	char *id = NULL;
+
+	if (fmt) {
+		va_list ap;
+
+		va_start(ap, fmt);
+		id = talloc_vasprintf(fi, fmt, ap);
+		va_end(ap);
+
+		if (!osmo_identifier_valid(id)) {
+			LOGP(DLGLOBAL, LOGL_ERROR,
+			     "Attempting to set illegal id for FSM instance of type '%s': %s\n",
+			     fi->fsm->name, osmo_quote_str(id, -1));
+			talloc_free(id);
+			return -EINVAL;
+		}
+	}
+
+	if (fi->id)
+		talloc_free((char*)fi->id);
+	fi->id = id;
+
+	update_name(fi);
+	return 0;
+}
+
 /*! allocate a new instance of a specified FSM
  *  \param[in] fsm Descriptor of the FSM
  *  \param[in] ctx talloc context from which to allocate memory
  *  \param[in] priv private data reference store in fsm instance
  *  \param[in] log_level The log level for events of this FSM
+ *  \param[in] id The name/ID of the FSM instance
  *  \returns newly-allocated, initialized and registered FSM instance
  */
 struct osmo_fsm_inst *osmo_fsm_inst_alloc(struct osmo_fsm *fsm, void *ctx, void *priv,
@@ -210,27 +286,10 @@ struct osmo_fsm_inst *osmo_fsm_inst_alloc(struct osmo_fsm *fsm, void *ctx, void
 	fi->priv = priv;
 	fi->log_level = log_level;
 	osmo_timer_setup(&fi->timer, fsm_tmr_cb, fi);
-	if (id) {
-		if (!osmo_identifier_valid(id)) {
-			LOGP(DLGLOBAL, LOGL_ERROR, "Attempting to allocate FSM instance of type '%s'"
-			     " with illegal identifier '%s'\n", fsm->name, id);
+
+	if (osmo_fsm_inst_update_id(fi, id) < 0) {
 			talloc_free(fi);
 			return NULL;
-		}
-		fi->id = talloc_strdup(fi, id);
-	}
-
-	if (!fsm_log_addr) {
-		if (id)
-			fi->name = talloc_asprintf(fi, "%s(%s)", fsm->name, id);
-		else
-			fi->name = talloc_asprintf(fi, "%s", fsm->name);
-	} else {
-		if (id)
-			fi->name = talloc_asprintf(fi, "%s(%s)[%p]", fsm->name,
-						   id, fi);
-		else
-			fi->name = talloc_asprintf(fi, "%s[%p]", fsm->name, fi);
 	}
 
 	INIT_LLIST_HEAD(&fi->proc.children);
@@ -269,13 +328,51 @@ struct osmo_fsm_inst *osmo_fsm_inst_alloc_child(struct osmo_fsm *fsm,
 
 	LOGPFSM(fi, "is child of %s\n", osmo_fsm_inst_name(parent));
 
-	fi->proc.parent = parent;
-	fi->proc.parent_term_event = parent_term_event;
-	llist_add(&fi->proc.child, &parent->proc.children);
+	osmo_fsm_inst_change_parent(fi, parent, parent_term_event);
 
 	return fi;
 }
 
+/*! unlink child FSM from its parent FSM.
+ *  \param[in] fi Descriptor of the child FSM to unlink.
+ *  \param[in] ctx New talloc context
+ *
+ * Never call this function from the cleanup callback, because at that time
+ * the child FSMs will already be terminated. If unlinking should be performed
+ * on FSM termination, use the grace callback instead. */
+void osmo_fsm_inst_unlink_parent(struct osmo_fsm_inst *fi, void *ctx)
+{
+	if (fi->proc.parent) {
+		talloc_steal(ctx, fi);
+		fi->proc.parent = NULL;
+		fi->proc.parent_term_event = 0;
+		llist_del(&fi->proc.child);
+	}
+}
+
+/*! change parent instance of an FSM.
+ *  \param[in] fi Descriptor of the to-be-allocated FSM.
+ *  \param[in] new_parent New parent FSM instance.
+ *  \param[in] new_parent_term_event Event to be sent to parent when terminating.
+ *
+ * Never call this function from the cleanup callback!
+ * (see also osmo_fsm_inst_unlink_parent()).*/
+void osmo_fsm_inst_change_parent(struct osmo_fsm_inst *fi,
+				 struct osmo_fsm_inst *new_parent,
+				 uint32_t new_parent_term_event)
+{
+	/* Make sure a possibly existing old parent is unlinked first
+	 * (new_parent can be NULL) */
+	osmo_fsm_inst_unlink_parent(fi, new_parent);
+
+	/* Add new parent */
+	if (new_parent) {
+		fi->proc.parent = new_parent;
+		fi->proc.parent_term_event = new_parent_term_event;
+		llist_add(&fi->proc.child, &new_parent->proc.children);
+	}
+}
+
 /*! delete a given instance of a FSM
  *  \param[in] fsm The FSM to be un-registered and deleted
  */
@@ -332,15 +429,56 @@ const char *osmo_fsm_state_name(struct osmo_fsm *fsm, uint32_t state)
 		return fsm->states[state].name;
 }
 
+static int state_chg(struct osmo_fsm_inst *fi, uint32_t new_state,
+		     bool keep_timer, unsigned long timeout_secs, int T,
+		     const char *file, int line)
+{
+	struct osmo_fsm *fsm = fi->fsm;
+	uint32_t old_state = fi->state;
+	const struct osmo_fsm_state *st = &fsm->states[fi->state];
+
+	/* validate if new_state is a valid state */
+	if (!(st->out_state_mask & (1 << new_state))) {
+		LOGPFSMLSRC(fi, LOGL_ERROR, file, line,
+			    "transition to state %s not permitted!\n",
+			    osmo_fsm_state_name(fsm, new_state));
+		return -EPERM;
+	}
+
+	if (!keep_timer) {
+		/* delete the old timer */
+		osmo_timer_del(&fi->timer);
+	}
+
+	if (st->onleave)
+		st->onleave(fi, new_state);
+
+	LOGPFSMSRC(fi, file, line, "state_chg to %s\n",
+		   osmo_fsm_state_name(fsm, new_state));
+	fi->state = new_state;
+	st = &fsm->states[new_state];
+
+	if (!keep_timer && timeout_secs) {
+		fi->T = T;
+		osmo_timer_schedule(&fi->timer, timeout_secs, 0);
+	}
+
+	/* Call 'onenter' last, user might terminate FSM from there */
+	if (st->onenter)
+		st->onenter(fi, old_state);
+
+	return 0;
+}
+
 /*! perform a state change of the given FSM instance
  *
  *  Best invoke via the osmo_fsm_inst_state_chg() macro which logs the source
  *  file where the state change was effected. Alternatively, you may pass \a
  *  file as NULL to use the normal file/line indication instead.
  *
- *  All changes to the FSM instance state must be made via this
+ *  All changes to the FSM instance state must be made via an osmo_fsm_inst_state_chg_*
  *  function.  It verifies that the existing state actually permits a
- *  transiiton to new_state.
+ *  transition to new_state.
  *
  *  timeout_secs and T are optional parameters, and only have any effect
  *  if timeout_secs is not 0.  If the timeout function is used, then the
@@ -360,39 +498,32 @@ int _osmo_fsm_inst_state_chg(struct osmo_fsm_inst *fi, uint32_t new_state,
 			     unsigned long timeout_secs, int T,
 			     const char *file, int line)
 {
-	struct osmo_fsm *fsm = fi->fsm;
-	uint32_t old_state = fi->state;
-	const struct osmo_fsm_state *st = &fsm->states[fi->state];
+	return state_chg(fi, new_state, false, timeout_secs, T, file, line);
+}
 
-	/* validate if new_state is a valid state */
-	if (!(st->out_state_mask & (1 << new_state))) {
-		LOGPFSMLSRC(fi, LOGL_ERROR, file, line,
-			    "transition to state %s not permitted!\n",
-			    osmo_fsm_state_name(fsm, new_state));
-		return -EPERM;
-	}
-
-	/* delete the old timer */
-	osmo_timer_del(&fi->timer);
-
-	if (st->onleave)
-		st->onleave(fi, new_state);
-
-	LOGPFSMSRC(fi, file, line, "state_chg to %s\n",
-		   osmo_fsm_state_name(fsm, new_state));
-	fi->state = new_state;
-	st = &fsm->states[new_state];
-
-	if (timeout_secs) {
-		fi->T = T;
-		osmo_timer_schedule(&fi->timer, timeout_secs, 0);
-	}
-
-	/* Call 'onenter' last, user might terminate FSM from there */
-	if (st->onenter)
-		st->onenter(fi, old_state);
-
-	return 0;
+/*! perform a state change while keeping the current timer running.
+ *
+ *  This is useful to keep a timeout across several states (without having to round the
+ *  remaining time to seconds).
+ *
+ *  Best invoke via the osmo_fsm_inst_state_chg_keep_timer() macro which logs the source
+ *  file where the state change was effected. Alternatively, you may pass \a
+ *  file as NULL to use the normal file/line indication instead.
+ *
+ *  All changes to the FSM instance state must be made via an osmo_fsm_inst_state_chg_*
+ *  function.  It verifies that the existing state actually permits a
+ *  transition to new_state.
+ *
+ *  \param[in] fi FSM instance whose state is to change
+ *  \param[in] new_state The new state into which we should change
+ *  \param[in] file Calling source file (from osmo_fsm_inst_state_chg macro)
+ *  \param[in] line Calling source line (from osmo_fsm_inst_state_chg macro)
+ *  \returns 0 on success; negative on error
+ */
+int _osmo_fsm_inst_state_chg_keep_timer(struct osmo_fsm_inst *fi, uint32_t new_state,
+					const char *file, int line)
+{
+	return state_chg(fi, new_state, true, 0, 0, file, line);
 }
 
 /*! dispatch an event to an osmocom finite state machine instance
@@ -444,7 +575,9 @@ int _osmo_fsm_inst_dispatch(struct osmo_fsm_inst *fi, uint32_t event, void *data
 			    osmo_fsm_event_name(fsm, event));
 		return -1;
 	}
-	fs->action(fi, event, data);
+
+	if (fs->action)
+		fs->action(fi, event, data);
 
 	return 0;
 }
@@ -475,15 +608,20 @@ void _osmo_fsm_inst_term(struct osmo_fsm_inst *fi,
 	LOGPFSMSRC(fi, file, line, "Terminating (cause = %s)\n",
 		   osmo_fsm_term_cause_name(cause));
 
+	/* graceful exit (optional) */
+	if (fi->fsm->pre_term)
+		fi->fsm->pre_term(fi, cause);
+
 	_osmo_fsm_inst_term_children(fi, OSMO_FSM_TERM_PARENT, NULL,
 				     file, line);
 
 	/* delete ourselves from the parent */
 	parent = fi->proc.parent;
-	if (parent)
+	if (parent) {
 		LOGPFSMSRC(fi, file, line, "Removing from parent %s\n",
 			   osmo_fsm_inst_name(parent));
-	llist_del(&fi->proc.child);
+		llist_del(&fi->proc.child);
+	}
 
 	/* call destructor / clean-up function */
 	if (fi->fsm->cleanup)
