@@ -2,10 +2,13 @@
  * GPRS Networks Service (NS) messages on the Gb interface.
  * 3GPP TS 08.16 version 8.0.1 Release 1999 / ETSI TS 101 299 V8.0.1 (2002-05). */
 /*
- * (C) 2009-2012 by Harald Welte <laforge@gnumonks.org>
+ * (C) 2009-2017 by Harald Welte <laforge@gnumonks.org>
+ * (C) 2016-2017 sysmocom - s.f.m.c. GmbH
  *
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -1702,4 +1705,26 @@ void gprs_ns_set_log_ss(int ss)
 	DNS = ss;
 }
 
+/*! Append the nsvc state to a talloc string
+ * \param s The string to append to (allocated with talloc)
+ * \param[in] nsvc The NS-VC to print the state of
+ * \returns The new string with state information appended to it
+ *
+ * This function will append a comma-separated state of the NS-VC to the
+ * string. The string needs to be allocated with talloc (e.g. talloc_strdup)
+ */
+char *gprs_nsvc_state_append(char *s, struct gprs_nsvc *nsvc)
+{
+	s = talloc_asprintf_append(s,
+		"%u,%u,%s,%s,%s,%s,%s\n",
+		nsvc->nsei, nsvc->nsvci,
+		NS_DESC_A(nsvc->state),
+		NS_DESC_B(nsvc->state),
+		nsvc->remote_end_is_sgsn ? "SGSN" : "BSS",
+		NS_DESC_A(nsvc->remote_state),
+		NS_DESC_B(nsvc->remote_state));
+
+	return s;
+}
+
 /*! @} */
