@@ -7,6 +7,8 @@
  *
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -28,6 +30,7 @@
 #include <string.h>
 #include <stdbool.h>
 #include <inttypes.h>
+#include <ctype.h>
 
 #include <osmocom/core/utils.h>
 #include <osmocom/core/byteswap.h>
@@ -38,6 +41,7 @@
 #include <osmocom/gsm/gsm0502.h>
 #include <osmocom/gsm/gsm_utils.h>
 #include <osmocom/gsm/protocol/gsm_04_08.h>
+#include <osmocom/gsm/protocol/gsm_04_80.h>
 #include <osmocom/gsm/protocol/gsm_08_58.h>
 #include <osmocom/gsm/protocol/gsm_04_08_gprs.h>
 
@@ -178,6 +182,19 @@ const char *rr_cause_name(uint8_t cause)
 	return get_value_string(rr_cause_names, cause);
 }
 
+/*! Return MCC-MNC-LAC-RAC as string, in a static buffer.
+ * \param[in] rai  RAI to encode.
+ * \returns Static string buffer.
+ */
+const char *osmo_rai_name(const struct gprs_ra_id *rai)
+{
+	static char buf[32];
+	snprintf(buf, sizeof(buf), "%s-%s-%u-%u",
+		 osmo_mcc_name(rai->mcc), osmo_mnc_name(rai->mnc, rai->mnc_3_digits), rai->lac,
+		 rai->rac);
+	return buf;
+}
+
 /* FIXME: convert to value_string */
 static const char *cc_state_names[32] = {
 	"NULL",
@@ -416,15 +433,6 @@ const char *gsm48_mi_type_name(uint8_t mi)
 	return get_value_string(mi_type_names, mi);
 }
 
-static void to_bcd(uint8_t *bcd, uint16_t val)
-{
-	bcd[2] = val % 10;
-	val = val / 10;
-	bcd[1] = val % 10;
-	val = val / 10;
-	bcd[0] = val % 10;
-}
-
 /*! Checks is particular message is cipherable in A/Gb mode according to
  *         3GPP TS 24.008 § 4.7.1.2
  *  \param[in] hdr Message header
@@ -449,64 +457,61 @@ bool gsm48_hdr_gmm_cipherable(const struct gsm48_hdr *hdr)
 	}
 }
 
-/* Convert MCC + MNC to BCD representation
- * \param[out] bcd_dst caller-allocated memory for output
- * \param[in] mcc Mobile Country Code
- * \param[in] mnc Mobile Network Code
- *
- * Convert given mcc and mnc to BCD and write to *bcd_dst, which must be an
- * allocated buffer of (at least) 3 bytes length. */
+/* Convert MCC + MNC to BCD representation, legacy implementation.
+ * Instead use osmo_plmn_to_bcd(), which is also capable of converting
+ * 3-digit MNC that have leading zeros. For parameters, also see there. */
 void gsm48_mcc_mnc_to_bcd(uint8_t *bcd_dst, uint16_t mcc, uint16_t mnc)
 {
-	uint8_t bcd[3];
-
-	to_bcd(bcd, mcc);
-	bcd_dst[0] = bcd[0] | (bcd[1] << 4);
-	bcd_dst[1] = bcd[2];
-
-	to_bcd(bcd, mnc);
-	/* FIXME: do we need three-digit MNC? See Table 10.5.3 */
-	if (mnc > 99) {
-		bcd_dst[1] |= bcd[2] << 4;
-		bcd_dst[2] = bcd[0] | (bcd[1] << 4);
-	} else {
-		bcd_dst[1] |= 0xf << 4;
-		bcd_dst[2] = bcd[1] | (bcd[2] << 4);
-	}
+	const struct osmo_plmn_id plmn = {
+		.mcc = mcc,
+		.mnc = mnc,
+		.mnc_3_digits = false,
+	};
+	osmo_plmn_to_bcd(bcd_dst, &plmn);
 }
 
-/* Convert given 3-byte BCD buffer to integers and write results to *mcc and
- * *mnc. The first three BCD digits result in the MCC and the remaining ones in
- * the MNC. */
+/* Convert given 3-byte BCD buffer to integers, legacy implementation.
+ * Instead use osmo_plmn_from_bcd(), which is also capable of converting
+ * 3-digit MNC that have leading zeros. For parameters, also see there. */
 void gsm48_mcc_mnc_from_bcd(uint8_t *bcd_src, uint16_t *mcc, uint16_t *mnc)
 {
-	*mcc = (bcd_src[0] & 0x0f) * 100
-	     + (bcd_src[0] >> 4) * 10
-	     + (bcd_src[1] & 0x0f);
-
-	if ((bcd_src[1] & 0xf0) == 0xf0) {
-		*mnc = (bcd_src[2] & 0x0f) * 10
-		     + (bcd_src[2] >> 4);
-	} else {
-		*mnc = (bcd_src[2] & 0x0f) * 100
-		     + (bcd_src[2] >> 4) * 10
-		     + (bcd_src[1] >> 4);
-	}
+	struct osmo_plmn_id plmn;
+	osmo_plmn_from_bcd(bcd_src, &plmn);
+	*mcc = plmn.mcc;
+	*mnc = plmn.mnc;
 }
 
-/*! Encode TS 04.08 Location Area Identifier
- *  \param[out] caller-provided memory for output
+/*! Encode TS 04.08 Location Area Identifier, legacy implementation.
+ * Instead use gsm48_generate_lai2(), which is capable of three-digit MNC with leading zeros.
+ *  \param[out] lai48 caller-provided memory for output
  *  \param[in] mcc Mobile Country Code
  *  \param[in] mnc Mobile Network Code
  *  \param[in] lac Location Area Code */
 void gsm48_generate_lai(struct gsm48_loc_area_id *lai48, uint16_t mcc,
 			uint16_t mnc, uint16_t lac)
 {
-	gsm48_mcc_mnc_to_bcd(&lai48->digits[0], mcc, mnc);
-	lai48->lac = osmo_htons(lac);
+	const struct osmo_location_area_id lai = {
+		.plmn = {
+			.mcc = mcc,
+			.mnc = mnc,
+			.mnc_3_digits = false,
+		},
+		.lac = lac,
+	};
+	gsm48_generate_lai2(lai48, &lai);
 }
 
-/*! Decode TS 04.08 Location Area Identifier
+/*! Encode TS 04.08 Location Area Identifier.
+ *  \param[out] lai48 caller-provided memory for output.
+ *  \param[in] lai input of MCC-MNC-LAC. */
+void gsm48_generate_lai2(struct gsm48_loc_area_id *lai48, const struct osmo_location_area_id *lai)
+{
+	osmo_plmn_to_bcd(&lai48->digits[0], &lai->plmn);
+	lai48->lac = osmo_htons(lai->lac);
+}
+
+/*! Decode TS 04.08 Location Area Identifier, legacy implementation.
+ * Instead use gsm48_decode_lai2(), which is capable of three-digit MNC with leading zeros.
  *  \param[in] Location Area Identifier (encoded)
  *  \param[out] mcc Mobile Country Code
  *  \param[out] mnc Mobile Network Code
@@ -517,11 +522,25 @@ void gsm48_generate_lai(struct gsm48_loc_area_id *lai48, uint16_t mcc,
 int gsm48_decode_lai(struct gsm48_loc_area_id *lai, uint16_t *mcc,
 		     uint16_t *mnc, uint16_t *lac)
 {
-	gsm48_mcc_mnc_from_bcd(&lai->digits[0], mcc, mnc);
-	*lac = osmo_ntohs(lai->lac);
+	struct osmo_location_area_id decoded;
+	gsm48_decode_lai2(lai, &decoded);
+	*mcc = decoded.plmn.mcc;
+	*mnc = decoded.plmn.mnc;
+	*lac = decoded.lac;
 	return 0;
 }
 
+/*! Decode TS 04.08 Location Area Identifier.
+ *  \param[in] Location Area Identifier (encoded).
+ *  \param[out] decoded Target buffer to write decoded values of MCC-MNC-LAC.
+ *
+ * Attention: this function returns true integers, not hex! */
+void gsm48_decode_lai2(const struct gsm48_loc_area_id *lai, struct osmo_location_area_id *decoded)
+{
+	osmo_plmn_from_bcd(&lai->digits[0], &decoded->plmn);
+	decoded->lac = osmo_ntohs(lai->lac);
+}
+
 /*! Set DTX mode in Cell Options IE (3GPP TS 44.018)
  *  \param[in] op Cell Options structure in which DTX parameters will be set
  *  \param[in] full Mode for full-rate channels
@@ -582,37 +601,40 @@ int gsm48_generate_mid_from_tmsi(uint8_t *buf, uint32_t tmsi)
 	return 7;
 }
 
+/*! Generate TS 24.008 §10.5.1.4 Mobile ID
+ *  \param[out] buf Caller-provided output buffer
+ *  \param[in] id Identity to be encoded
+ *  \returns number of bytes used in \a buf */
+uint8_t gsm48_generate_mid(uint8_t *buf, const char *id, uint8_t mi_type)
+{
+	uint8_t length = strnlen(id, 255), i, off = 0, odd = (length & 1) == 1;
+
+	buf[0] = GSM48_IE_MOBILE_ID;
+	buf[2] = osmo_char2bcd(id[0]) << 4 | mi_type | (odd << 3);
+
+	/* if the length is even we will fill half of the last octet */
+	buf[1] = (length + (odd ? 1 : 2)) >> 1;
+
+	for (i = 1; i < buf[1]; ++i) {
+		uint8_t upper, lower = osmo_char2bcd(id[++off]);
+		if (!odd && off + 1 == length)
+			upper = 0x0f;
+		else
+			upper = osmo_char2bcd(id[++off]) & 0x0f;
+
+		buf[2 + i] = (upper << 4) | lower;
+	}
+
+	return 2 + buf[1];
+}
+
 /*! Generate TS 04.08 Mobile ID from IMSI
  *  \param[out] buf Caller-provided output buffer
  *  \param[in] imsi IMSI to be encoded
  *  \returns number of bytes used in \a buf */
 int gsm48_generate_mid_from_imsi(uint8_t *buf, const char *imsi)
 {
-	unsigned int length = strlen(imsi), i, off = 0;
-	uint8_t odd = (length & 0x1) == 1;
-
-	buf[0] = GSM48_IE_MOBILE_ID;
-	buf[2] = osmo_char2bcd(imsi[0]) << 4 | GSM_MI_TYPE_IMSI | (odd << 3);
-
-	/* if the length is even we will fill half of the last octet */
-	if (odd)
-		buf[1] = (length + 1) >> 1;
-	else
-		buf[1] = (length + 2) >> 1;
-
-	for (i = 1; i < buf[1]; ++i) {
-		uint8_t lower, upper;
-
-		lower = osmo_char2bcd(imsi[++off]);
-		if (!odd && off + 1 == length)
-			upper = 0x0f;
-		else
-			upper = osmo_char2bcd(imsi[++off]) & 0x0f;
-
-		buf[2 + i] = (upper << 4) | lower;
-	}
-
-	return 2 + buf[1];
+	return gsm48_generate_mid(buf, imsi, GSM_MI_TYPE_IMSI);
 }
 
 /*! Convert TS 04.08 Mobile Identity (10.5.1.4) to string
@@ -677,43 +699,46 @@ void gsm48_parse_ra(struct gprs_ra_id *raid, const uint8_t *buf)
 	if ((buf[1] >> 4) == 0xf) {
 		raid->mnc = (buf[2] & 0xf) * 10;
 		raid->mnc += (buf[2] >> 4) * 1;
+		raid->mnc_3_digits = false;
 	} else {
 		raid->mnc = (buf[2] & 0xf) * 100;
 		raid->mnc += (buf[2] >> 4) * 10;
 		raid->mnc += (buf[1] >> 4) * 1;
+		raid->mnc_3_digits = true;
 	}
 
 	raid->lac = osmo_load16be(buf + 3);
 	raid->rac = buf[5];
 }
 
+/*! Encode a 3GPP TS 24.008 § 10.5.5.15 Routing area identification
+ *  \param[out] out Caller-provided packed struct
+ *  \param[in] raid Routing Area ID to be encoded
+ */
+void gsm48_encode_ra(struct gsm48_ra_id *out, const struct gprs_ra_id *raid)
+{
+	out->lac = osmo_htons(raid->lac);
+	out->rac = raid->rac;
+
+	out->digits[0] = ((raid->mcc / 100) % 10) | (((raid->mcc / 10) % 10) << 4);
+	out->digits[1] = raid->mcc % 10;
+
+	if (raid->mnc < 100 && !raid->mnc_3_digits) {
+		out->digits[1] |= 0xf0;
+		out->digits[2] = ((raid->mnc / 10) % 10) | ((raid->mnc % 10) << 4);
+	} else {
+		out->digits[1] |= (raid->mnc % 10) << 4;
+		out->digits[2] = ((raid->mnc / 100) % 10) | (((raid->mnc / 10) % 10) << 4);
+	}
+}
+
 /*! Encode a TS 04.08 Routing Area Identifier
  *  \param[out] buf Caller-provided output buffer of 6 bytes
  *  \param[in] raid Routing Area ID to be encoded
  *  \returns number of bytes used in \a buf */
 int gsm48_construct_ra(uint8_t *buf, const struct gprs_ra_id *raid)
 {
-	uint16_t mcc = raid->mcc;
-	uint16_t mnc = raid->mnc;
-	uint16_t _lac;
-
-	buf[0] = ((mcc / 100) % 10) | (((mcc / 10) % 10) << 4);
-	buf[1] = (mcc % 10);
-
-	/* I wonder who came up with the stupidity of encoding the MNC
-	 * differently depending on how many digits its decimal number has! */
-	if (mnc < 100) {
-		buf[1] |= 0xf0;
-		buf[2] = ((mnc / 10) % 10) | ((mnc % 10) << 4);
-	} else {
-		buf[1] |= (mnc % 10) << 4;
-		buf[2] = ((mnc / 100) % 10) | (((mnc / 10) % 10) << 4);
-	}
-
-	_lac = osmo_htons(raid->lac);
-	memcpy(buf + 3, &_lac, 2);
-
-	buf[5] = raid->rac;
+	gsm48_encode_ra((struct gsm48_ra_id *)buf, raid);
 
 	return 6;
 }
@@ -737,21 +762,20 @@ int gsm48_number_of_paging_subchannels(struct gsm48_control_channel_descr *chan_
 
 /*! TS 04.08 Protocol Descriptor names */
 const struct value_string gsm48_pdisc_names[] = {
-	OSMO_VALUE_STRING(GSM48_PDISC_GROUP_CC),
-	OSMO_VALUE_STRING(GSM48_PDISC_BCAST_CC),
-	OSMO_VALUE_STRING(GSM48_PDISC_PDSS1),
-	OSMO_VALUE_STRING(GSM48_PDISC_CC),
-	OSMO_VALUE_STRING(GSM48_PDISC_PDSS2),
-	OSMO_VALUE_STRING(GSM48_PDISC_MM),
-	OSMO_VALUE_STRING(GSM48_PDISC_RR),
-	OSMO_VALUE_STRING(GSM48_PDISC_MM_GPRS),
-	OSMO_VALUE_STRING(GSM48_PDISC_SMS),
-	OSMO_VALUE_STRING(GSM48_PDISC_SM_GPRS),
-	OSMO_VALUE_STRING(GSM48_PDISC_NC_SS),
-	OSMO_VALUE_STRING(GSM48_PDISC_LOC),
-	OSMO_VALUE_STRING(GSM48_PDISC_EXTEND),
-	OSMO_VALUE_STRING(GSM48_PDISC_MASK),
-	OSMO_VALUE_STRING(GSM48_PDISC_USSD),
+	{ GSM48_PDISC_GROUP_CC,		"VGCC" },
+	{ GSM48_PDISC_BCAST_CC,		"VBCC" },
+	{ GSM48_PDISC_PDSS1,		"PDSS1" },
+	{ GSM48_PDISC_CC,		"CC" },
+	{ GSM48_PDISC_PDSS2,		"PDSS2" },
+	{ GSM48_PDISC_MM,		"MM" },
+	{ GSM48_PDISC_RR,		"RR" },
+	{ GSM48_PDISC_MM_GPRS,		"GMM" },
+	{ GSM48_PDISC_SMS,		"SMS" },
+	{ GSM48_PDISC_SM_GPRS,		"SM" },
+	{ GSM48_PDISC_NC_SS,		"NCSS" },
+	{ GSM48_PDISC_LOC,		"LCS" },
+	{ GSM48_PDISC_EXTEND,		"EXTD" },
+	{ GSM48_PDISC_MASK,		"MASK" },
 	{ 0, NULL }
 };
 
@@ -921,6 +945,14 @@ const struct value_string gsm48_cc_msgtype_names[] = {
 	{ 0, NULL }
 };
 
+/*! TS 04.80, section 3.4 Messages for supplementary services control */
+const struct value_string gsm48_nc_ss_msgtype_names[] = {
+	OSMO_VALUE_STRING(GSM0480_MTYPE_RELEASE_COMPLETE),
+	OSMO_VALUE_STRING(GSM0480_MTYPE_FACILITY),
+	OSMO_VALUE_STRING(GSM0480_MTYPE_REGISTER),
+	{ 0, NULL }
+};
+
 /*! Compose a string naming the message type for given protocol.
  * If the message type string is known, return the message type name, otherwise
  * return "<protocol discriminator name>:<message type in hex>".
@@ -943,6 +975,9 @@ const char *gsm48_pdisc_msgtype_name(uint8_t pdisc, uint8_t msg_type)
 	case GSM48_PDISC_CC:
 		msgt_names = gsm48_cc_msgtype_names;
 		break;
+	case GSM48_PDISC_NC_SS:
+		msgt_names = gsm48_nc_ss_msgtype_names;
+		break;
 	default:
 		msgt_names = NULL;
 		break;
@@ -956,4 +991,37 @@ const char *gsm48_pdisc_msgtype_name(uint8_t pdisc, uint8_t msg_type)
 	return namebuf;
 }
 
+const struct value_string gsm48_reject_value_names[] = {
+	 { GSM48_REJECT_IMSI_UNKNOWN_IN_HLR, "IMSI_UNKNOWN_IN_HLR" },
+	 { GSM48_REJECT_ILLEGAL_MS, "ILLEGAL_MS" },
+	 { GSM48_REJECT_IMSI_UNKNOWN_IN_VLR, "IMSI_UNKNOWN_IN_VLR" },
+	 { GSM48_REJECT_IMEI_NOT_ACCEPTED, "IMEI_NOT_ACCEPTED" },
+	 { GSM48_REJECT_ILLEGAL_ME, "ILLEGAL_ME" },
+	 { GSM48_REJECT_PLMN_NOT_ALLOWED, "PLMN_NOT_ALLOWED" },
+	 { GSM48_REJECT_LOC_NOT_ALLOWED, "LOC_NOT_ALLOWED" },
+	 { GSM48_REJECT_ROAMING_NOT_ALLOWED, "ROAMING_NOT_ALLOWED" },
+	 { GSM48_REJECT_NETWORK_FAILURE, "NETWORK_FAILURE" },
+	 { GSM48_REJECT_SYNCH_FAILURE, "SYNCH_FAILURE" },
+	 { GSM48_REJECT_CONGESTION, "CONGESTION" },
+	 { GSM48_REJECT_SRV_OPT_NOT_SUPPORTED, "SRV_OPT_NOT_SUPPORTED" },
+	 { GSM48_REJECT_RQD_SRV_OPT_NOT_SUPPORTED, "RQD_SRV_OPT_NOT_SUPPORTED" },
+	 { GSM48_REJECT_SRV_OPT_TMP_OUT_OF_ORDER, "SRV_OPT_TMP_OUT_OF_ORDER" },
+	 { GSM48_REJECT_CALL_CAN_NOT_BE_IDENTIFIED, "CALL_CAN_NOT_BE_IDENTIFIED" },
+	 { GSM48_REJECT_INCORRECT_MESSAGE, "INCORRECT_MESSAGE" },
+	 { GSM48_REJECT_INVALID_MANDANTORY_INF, "INVALID_MANDANTORY_INF" },
+	 { GSM48_REJECT_MSG_TYPE_NOT_IMPLEMENTED, "MSG_TYPE_NOT_IMPLEMENTED" },
+	 { GSM48_REJECT_MSG_TYPE_NOT_COMPATIBLE, "MSG_TYPE_NOT_COMPATIBLE" },
+	 { GSM48_REJECT_INF_ELEME_NOT_IMPLEMENTED, "INF_ELEME_NOT_IMPLEMENTED" },
+	 { GSM48_REJECT_CONDTIONAL_IE_ERROR, "CONDTIONAL_IE_ERROR" },
+	 { GSM48_REJECT_MSG_NOT_COMPATIBLE, "MSG_NOT_COMPATIBLE" },
+	 { GSM48_REJECT_PROTOCOL_ERROR, "PROTOCOL_ERROR" },
+	 { GSM48_REJECT_GPRS_NOT_ALLOWED, "GPRS_NOT_ALLOWED" },
+	 { GSM48_REJECT_SERVICES_NOT_ALLOWED, "SERVICES_NOT_ALLOWED" },
+	 { GSM48_REJECT_MS_IDENTITY_NOT_DERVIVABLE, "MS_IDENTITY_NOT_DERVIVABLE" },
+	 { GSM48_REJECT_IMPLICITLY_DETACHED, "IMPLICITLY_DETACHED" },
+	 { GSM48_REJECT_GPRS_NOT_ALLOWED_IN_PLMN, "GPRS_NOT_ALLOWED_IN_PLMN" },
+	 { GSM48_REJECT_MSC_TMP_NOT_REACHABLE, "MSC_TMP_NOT_REACHABLE" },
+	 { 0, NULL }
+};
+
 /*! @} */
