@@ -1,11 +1,14 @@
 /*! \file lapdm.c
  * GSM LAPDm (TS 04.06) implementation. */
 /*
- * (C) 2010-2011 by Harald Welte <laforge@gnumonks.org>
+ * (C) 2010-2017 by Harald Welte <laforge@gnumonks.org>
  * (C) 2010-2011 by Andreas Eversberg <jolly@eversberg.eu>
+ * (C) 2014-2016 by sysmocom - s.f.m.c GmbH
  *
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -45,6 +48,8 @@
 #include <osmocom/gsm/protocol/gsm_04_08.h>
 #include <osmocom/gsm/protocol/gsm_08_58.h>
 
+#define LAPD_U_SABM 0x7
+
 /* TS 04.06 Figure 4 / Section 3.2 */
 #define LAPDm_LPD_NORMAL  0
 #define LAPDm_LPD_SMSCB	  1
@@ -166,8 +171,10 @@ void lapdm_entity_init(struct lapdm_entity *le, enum lapdm_mode mode, int t200)
 void lapdm_channel_init(struct lapdm_channel *lc, enum lapdm_mode mode)
 {
 	lapdm_entity_init(&lc->lapdm_acch, mode, 2);
+	lc->lapdm_acch.lapdm_ch = lc;
 	/* FIXME: this depends on chan type */
 	lapdm_entity_init(&lc->lapdm_dcch, mode, 1);
+	lc->lapdm_dcch.lapdm_ch = lc;
 }
 
 /*! flush and release all resoures in LAPDm entity */
@@ -346,12 +353,29 @@ static int l2_ph_data_conf(struct msgb *msg, struct lapdm_entity *le)
 	return le->l1_prim_cb(&pp.oph, le->l1_ctx);
 }
 
+/* Is a given msg_type "transparent" as per TS 48.058 Section 8.1 */
+static int rsl_is_transparent(uint8_t msg_type)
+{
+	switch (msg_type) {
+	case RSL_MT_DATA_IND:
+	case RSL_MT_UNIT_DATA_IND:
+		return 1;
+	case RSL_MT_DATA_REQ:
+	case RSL_MT_UNIT_DATA_REQ:
+		return 1;
+	default:
+		return 0;
+	}
+}
+
 /* Create RSLms various RSLms messages */
 static int send_rslms_rll_l3(uint8_t msg_type, struct lapdm_msg_ctx *mctx,
 			     struct msgb *msg)
 {
+	int transparent = rsl_is_transparent(msg_type);
+
 	/* Add the RSL + RLL header */
-	rsl_rll_push_l3(msg, msg_type, mctx->chan_nr, mctx->link_id, 1);
+	rsl_rll_push_l3(msg, msg_type, mctx->chan_nr, mctx->link_id, transparent);
 
 	/* send off the RSLms message to L3 */
 	return rslms_sendmsg(msg, mctx->dl->entity);
@@ -361,29 +385,28 @@ static int send_rslms_rll_l3(uint8_t msg_type, struct lapdm_msg_ctx *mctx,
 static int send_rslms_rll_l3_ui(struct lapdm_msg_ctx *mctx, struct msgb *msg)
 {
 	uint8_t l3_len = msg->tail - (uint8_t *)msgb_l3(msg);
-	struct abis_rsl_rll_hdr *rllh;
 
 	/* Add the RSL + RLL header */
 	msgb_tv16_push(msg, RSL_IE_L3_INFO, l3_len);
-	msgb_push(msg, 2 + 2);
+
+	/* Add two non-standard IEs carrying MS power and TA values for B4 (SACCH) */
+	if (mctx->lapdm_fmt == LAPDm_FMT_B4) {
+		msgb_tv_push(msg, RSL_IE_MS_POWER, mctx->tx_power_ind);
+		msgb_tv_push(msg, RSL_IE_TIMING_ADVANCE, mctx->ta_ind);
+	}
+
 	rsl_rll_push_hdr(msg, RSL_MT_UNIT_DATA_IND, mctx->chan_nr,
 		mctx->link_id, 1);
-	rllh = (struct abis_rsl_rll_hdr *)msgb_l2(msg);
 
-	rllh->data[0] = RSL_IE_TIMING_ADVANCE;
-	rllh->data[1] = mctx->ta_ind;
-
-	rllh->data[2] = RSL_IE_MS_POWER;
-	rllh->data[3] = mctx->tx_power_ind;
-	
 	return rslms_sendmsg(msg, mctx->dl->entity);
 }
 
 static int send_rll_simple(uint8_t msg_type, struct lapdm_msg_ctx *mctx)
 {
 	struct msgb *msg;
+	int transparent = rsl_is_transparent(msg_type);
 
-	msg = rsl_rll_simple(msg_type, mctx->chan_nr, mctx->link_id, 1);
+	msg = rsl_rll_simple(msg_type, mctx->chan_nr, mctx->link_id, transparent);
 
 	/* send off the RSLms message to L3 */
 	return rslms_sendmsg(msg, mctx->dl->entity);
@@ -394,7 +417,7 @@ static int rsl_rll_error(uint8_t cause, struct lapdm_msg_ctx *mctx)
 	struct msgb *msg;
 
 	LOGP(DLLAPD, LOGL_NOTICE, "sending MDL-ERROR-IND %d\n", cause);
-	msg = rsl_rll_simple(RSL_MT_ERROR_IND, mctx->chan_nr, mctx->link_id, 1);
+	msg = rsl_rll_simple(RSL_MT_ERROR_IND, mctx->chan_nr, mctx->link_id, 0);
 	msgb_tlv_put(msg, RSL_IE_RLM_CAUSE, 1, &cause);
 	return rslms_sendmsg(msg, mctx->dl->entity);
 }
@@ -411,11 +434,6 @@ static int send_rslms_dlsap(struct osmo_dlsap_prim *dp,
 
 	switch (OSMO_PRIM_HDR(&dp->oph)) {
 	case OSMO_PRIM(PRIM_DL_EST, PRIM_OP_INDICATION):
-		if (dp->oph.msg && dp->oph.msg->len == 0) {
-			/* omit L3 info by freeing message */
-			msgb_free(dp->oph.msg);
-			dp->oph.msg = NULL;
-		}
 		rll_msg = RSL_MT_EST_IND;
 		break;
 	case OSMO_PRIM(PRIM_DL_EST, PRIM_OP_CONFIRM):
@@ -519,6 +537,39 @@ static int update_pending_frames(struct lapd_msg_ctx *lctx)
 	return rc;
 }
 
+/* determine if receiving a given LAPDm message is not permitted */
+static int lapdm_rx_not_permitted(const struct lapdm_entity *le,
+				  const struct lapd_msg_ctx *lctx)
+{
+	/* we currently only implement SABM related checks here */
+	if (lctx->format != LAPD_FORM_U || lctx->s_u != LAPD_U_SABM)
+		return 0;
+
+	if (le->mode == LAPDM_MODE_BTS) {
+		if (le == &le->lapdm_ch->lapdm_acch) {
+			/* no contention resolution on SACCH */
+			if (lctx->length > 0)
+				return RLL_CAUSE_SABM_INFO_NOTALL;
+		} else {
+			switch (lctx->sapi) {
+			case 3:
+				/* SAPI3 doesn't support contention resolution */
+				if (lctx->length > 0)
+					return RLL_CAUSE_SABM_INFO_NOTALL;
+				break;
+			default:
+				break;
+			}
+		}
+	} else if (le->mode == LAPDM_MODE_MS) {
+		/* contention resolution (L3 present) is only sent by MS, but
+		 * never received by it */
+		if (lctx->length > 0)
+			return RLL_CAUSE_SABM_INFO_NOTALL;
+	}
+	return 0;
+}
+
 /* input into layer2 (from layer 1) */
 static int l2_ph_data_ind(struct msgb *msg, struct lapdm_entity *le,
 	uint8_t chan_nr, uint8_t link_id)
@@ -656,6 +707,13 @@ static int l2_ph_data_ind(struct msgb *msg, struct lapdm_entity *le,
 		}
 		/* store context for messages from lapd */
 		memcpy(&mctx.dl->mctx, &mctx, sizeof(mctx.dl->mctx));
+		rc =lapdm_rx_not_permitted(le, &lctx);
+		if (rc > 0) {
+			LOGP(DLLAPD, LOGL_NOTICE, "received message not permitted\n");
+			msgb_free(msg);
+			rsl_rll_error(rc, &mctx);
+			return -EINVAL;
+		}
 		/* send to LAPD */
 		rc = lapd_ph_data_ind(msg, &lctx);
 		break;
@@ -1083,7 +1141,7 @@ static int l2_ph_chan_conf(struct msgb *msg, struct lapdm_entity *le, uint32_t f
 	ref->t2 = tm.t2;
 	ref->t3_low = tm.t3 & 0x7;
 	ref->t3_high = tm.t3 >> 3;
-	
+
 	return rslms_sendmsg(msg, le);
 }
 
