@@ -5,6 +5,8 @@
  * (C) 2008 by Holger Hans Peter Freyther <zecke@selfish.org>
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -166,6 +168,11 @@ static const struct log_info_cat internal_cat[OSMO_NUM_DLIB] = {
 		.description = "libosmo-mgcp Media Gateway Control Protocol",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
+	[INT2IDX(DLJIBUF)] = {
+		.name = "DLJIBUF",
+		.description = "libosmo-netif Jitter Buffer",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
 };
 
 /*! descriptive string for each log level */
@@ -180,11 +187,11 @@ static const char *loglevel_descriptions[LOGLEVEL_DEFS+1] = {
 	NULL,
 };
 
-static void assert_loginfo(void)
+static void assert_loginfo(const char *src)
 {
 	if (!osmo_log_info) {
 		fprintf(stderr, "ERROR: osmo_log_info == NULL! "
-			"You must call log_init() before using logging!\n");
+			"You must call log_init() before using logging in %s()!\n", src);
 		OSMO_ASSERT(osmo_log_info);
 	}
 }
@@ -221,7 +228,7 @@ int log_parse_category(const char *category)
 {
 	int i;
 
-	assert_loginfo();
+	assert_loginfo(__func__);
 
 	for (i = 0; i < osmo_log_info->num_cat; ++i) {
 		if (osmo_log_info->cat[i].name == NULL)
@@ -246,7 +253,7 @@ void log_parse_category_mask(struct log_target* target, const char *_mask)
 	char *mask = strdup(_mask);
 	char *category_token = NULL;
 
-	assert_loginfo();
+	assert_loginfo(__func__);
 
 	/* Disable everything to enable it afterwards */
 	for (i = 0; i < osmo_log_info->num_cat; ++i)
@@ -296,6 +303,23 @@ static const char* color(int subsys)
 	return NULL;
 }
 
+static const struct value_string level_colors[] = {
+	{ LOGL_DEBUG, "\033[1;34m" },
+	{ LOGL_INFO, "\033[1;32m" },
+	{ LOGL_NOTICE, "\033[1;33m" },
+	{ LOGL_ERROR, "\033[1;31m" },
+	{ LOGL_FATAL, "\033[1;31m" },
+	{ 0, NULL }
+};
+
+static const char *level_color(int level)
+{
+	const char *c = get_value_string_or_null(level_colors, level);
+	if (!c)
+		return get_value_string(level_colors, LOGL_FATAL);
+	return c;
+}
+
 const char* log_category_name(int subsys)
 {
 	if (subsys < osmo_log_info->num_cat)
@@ -304,18 +328,27 @@ const char* log_category_name(int subsys)
 	return NULL;
 }
 
+static const char *const_basename(const char *path)
+{
+	const char *bn = strrchr(path, '/');
+	if (!bn || !bn[1])
+		return path;
+	return bn + 1;
+}
+
 static void _output(struct log_target *target, unsigned int subsys,
 		    unsigned int level, const char *file, int line, int cont,
 		    const char *format, va_list ap)
 {
 	char buf[4096];
 	int ret, len = 0, offset = 0, rem = sizeof(buf);
+	const char *c_subsys = NULL;
 
 	/* are we using color */
 	if (target->use_color) {
-		const char *c = color(subsys);
-		if (c) {
-			ret = snprintf(buf + offset, rem, "%s", c);
+		c_subsys = color(subsys);
+		if (c_subsys) {
+			ret = snprintf(buf + offset, rem, "%s", c_subsys);
 			if (ret < 0)
 				goto err;
 			OSMO_SNPRINTF_RET(ret, rem, offset, len);
@@ -323,6 +356,7 @@ static void _output(struct log_target *target, unsigned int subsys,
 	}
 	if (!cont) {
 		if (target->print_ext_timestamp) {
+#ifdef HAVE_LOCALTIME_R
 			struct tm tm;
 			struct timeval tv;
 			osmo_gettimeofday(&tv, NULL);
@@ -334,6 +368,7 @@ static void _output(struct log_target *target, unsigned int subsys,
 			if (ret < 0)
 				goto err;
 			OSMO_SNPRINTF_RET(ret, rem, offset, len);
+#endif
 		} else if (target->print_timestamp) {
 			char *timestr;
 			time_t tm;
@@ -346,29 +381,59 @@ static void _output(struct log_target *target, unsigned int subsys,
 			OSMO_SNPRINTF_RET(ret, rem, offset, len);
 		}
 		if (target->print_category) {
-			ret = snprintf(buf + offset, rem, "%s ", log_category_name(subsys));
+			ret = snprintf(buf + offset, rem, "%s%s%s%s ",
+				       target->use_color ? level_color(level) : "",
+				       log_category_name(subsys),
+				       target->use_color ? "\033[0;m" : "",
+				       c_subsys ? c_subsys : "");
 			if (ret < 0)
 				goto err;
 			OSMO_SNPRINTF_RET(ret, rem, offset, len);
 		}
-		if (target->print_filename) {
-			ret = snprintf(buf + offset, rem, "<%4.4x> %s:%d ",
-					subsys, file, line);
+		if (target->print_level) {
+			ret = snprintf(buf + offset, rem, "%s%s%s%s ",
+				       target->use_color ? level_color(level) : "",
+				       log_level_str(level),
+				       target->use_color ? "\033[0;m" : "",
+				       c_subsys ? c_subsys : "");
 			if (ret < 0)
 				goto err;
 			OSMO_SNPRINTF_RET(ret, rem, offset, len);
 		}
+		if (target->print_category_hex) {
+			ret = snprintf(buf + offset, rem, "<%4.4x> ", subsys);
+			if (ret < 0)
+				goto err;
+			OSMO_SNPRINTF_RET(ret, rem, offset, len);
+		}
+		switch (target->print_filename2) {
+		case LOG_FILENAME_NONE:
+			break;
+		case LOG_FILENAME_PATH:
+			ret = snprintf(buf + offset, rem, "%s:%d ", file, line);
+			if (ret < 0)
+				goto err;
+			OSMO_SNPRINTF_RET(ret, rem, offset, len);
+			break;
+		case LOG_FILENAME_BASENAME:
+			ret = snprintf(buf + offset, rem, "%s:%d ", const_basename(file), line);
+			if (ret < 0)
+				goto err;
+			OSMO_SNPRINTF_RET(ret, rem, offset, len);
+			break;
+		}
 	}
 	ret = vsnprintf(buf + offset, rem, format, ap);
 	if (ret < 0)
 		goto err;
 	OSMO_SNPRINTF_RET(ret, rem, offset, len);
 
-	ret = snprintf(buf + offset, rem, "%s",
-			target->use_color ? "\033[0;m" : "");
-	if (ret < 0)
-		goto err;
-	OSMO_SNPRINTF_RET(ret, rem, offset, len);
+	if (target->use_color) {
+		ret = snprintf(buf + offset, rem, "\033[0;m");
+		if (ret < 0)
+			goto err;
+		OSMO_SNPRINTF_RET(ret, rem, offset, len);
+	}
 err:
 	buf[sizeof(buf)-1] = '\0';
 	target->output(target, level, buf);
@@ -586,13 +651,30 @@ void log_set_print_extended_timestamp(struct log_target *target, int print_times
 	target->print_ext_timestamp = print_timestamp;
 }
 
-/*! Enable or disable printing of the filename while logging
+/*! Use log_set_print_filename2() instead.
+ * Call log_set_print_filename2() with LOG_FILENAME_PATH or LOG_FILENAME_NONE, *as well as* call
+ * log_set_print_category_hex() with the argument passed to this function. This is to mirror legacy
+ * behavior, which combined the category in hex with the filename. For example, if the category-hex
+ * output were no longer affected by log_set_print_filename(), many unit tests (in libosmocore as well as
+ * dependent projects) would fail since they expect the category to disappear along with the filename.
  *  \param[in] target Log target to be affected
  *  \param[in] print_filename Enable (1) or disable (0) filenames
  */
 void log_set_print_filename(struct log_target *target, int print_filename)
 {
-	target->print_filename = print_filename;
+	log_set_print_filename2(target, print_filename ? LOG_FILENAME_PATH : LOG_FILENAME_NONE);
+	log_set_print_category_hex(target, print_filename);
+}
+
+/*! Enable or disable printing of the filename while logging.
+ *  \param[in] target Log target to be affected.
+ *  \param[in] print_filename An LOG_FILENAME_* enum value.
+ * LOG_FILENAME_NONE omits the source file and line information from logs.
+ * LOG_FILENAME_PATH prints the entire source file path as passed to LOGP macros.
+ */
+void log_set_print_filename2(struct log_target *target, enum log_filename_type lft)
+{
+	target->print_filename2 = lft;
 }
 
 /*! Enable or disable printing of the category name
@@ -606,6 +688,26 @@ void log_set_print_category(struct log_target *target, int print_category)
 	target->print_category = print_category;
 }
 
+/*! Enable or disable printing of the category number in hex ('<000b>').
+ *  \param[in] target Log target to be affected.
+ *  \param[in] print_category_hex Enable (1) or disable (0) hex category.
+ */
+void log_set_print_category_hex(struct log_target *target, int print_category_hex)
+{
+	target->print_category_hex = print_category_hex;
+}
+
+/*! Enable or disable printing of the log level name.
+ *  \param[in] target Log target to be affected
+ *  \param[in] print_catname Enable (1) or disable (0) filenames
+ *
+ *  Print the log level name in front of every log message.
+ */
+void log_set_print_level(struct log_target *target, int print_level)
+{
+	target->print_level = (bool)print_level;
+}
+
 /*! Set the global log level for a given log target
  *  \param[in] target Log target to be affected
  *  \param[in] log_level New global log level
@@ -651,13 +753,13 @@ struct log_target *log_target_create(void)
 	struct log_target *target;
 	unsigned int i;
 
-	assert_loginfo();
+	assert_loginfo(__func__);
 
 	target = talloc_zero(tall_log_ctx, struct log_target);
 	if (!target)
 		return NULL;
 
-	target->categories = talloc_zero_array(target, 
+	target->categories = talloc_zero_array(target,
 						struct log_category,
 						osmo_log_info->num_cat);
 	if (!target->categories) {
@@ -677,7 +779,8 @@ struct log_target *log_target_create(void)
 	/* global settings */
 	target->use_color = 1;
 	target->print_timestamp = 0;
-	target->print_filename = 1;
+	target->print_filename2 = LOG_FILENAME_PATH;
+	target->print_category_hex = true;
 
 	/* global log level */
 	target->loglevel = 0;
@@ -743,11 +846,18 @@ struct log_target *log_target_find(int type, const char *fname)
 	llist_for_each_entry(tgt, &osmo_log_target_list, entry) {
 		if (tgt->type != type)
 			continue;
-		if (tgt->type == LOG_TGT_TYPE_FILE) {
+		switch (tgt->type) {
+		case LOG_TGT_TYPE_FILE:
 			if (!strcmp(fname, tgt->tgt_file.fname))
 				return tgt;
-		} else
+			break;
+		case LOG_TGT_TYPE_GSMTAP:
+			if (!strcmp(fname, tgt->tgt_gsmtap.hostname))
+				return tgt;
+			break;
+		default:
 			return tgt;
+		}
 	}
 	return NULL;
 }
@@ -822,10 +932,10 @@ const char *log_vty_command_string()
 {
 	struct log_info *info = osmo_log_info;
 	int len = 0, offset = 0, ret, i, rem;
-	int size = strlen("logging level () ()") + 1;
+	int size = strlen("logging level (all|) ()") + 1;
 	char *str;
 
-	assert_loginfo();
+	assert_loginfo(__func__);
 
 	for (i = 0; i < info->num_cat; i++) {
 		if (info->cat[i].name == NULL)
@@ -907,7 +1017,7 @@ const char *log_vty_command_description()
 		strlen(LOGGING_STR
 		       "Set the log level for a specified category\n") + 1;
 
-	assert_loginfo();
+	assert_loginfo(__func__);
 
 	for (i = 0; i < info->num_cat; i++) {
 		if (info->cat[i].name == NULL)
@@ -1026,7 +1136,7 @@ int log_check_level(int subsys, unsigned int level)
 {
 	struct log_target *tar;
 
-	assert_loginfo();
+	assert_loginfo(__func__);
 
 	subsys = map_subsys(subsys);
 
