@@ -3,6 +3,8 @@
  *
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -33,6 +35,7 @@
 #include <osmocom/core/select.h>
 #include <osmocom/core/socket.h>
 #include <osmocom/core/talloc.h>
+#include <osmocom/core/utils.h>
 
 #include <sys/ioctl.h>
 #include <sys/socket.h>
@@ -89,8 +92,11 @@ static int socket_helper(const struct addrinfo *rp, unsigned int flags)
 	int sfd, on = 1;
 
 	sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
-	if (sfd == -1)
+	if (sfd == -1) {
+		LOGP(DLGLOBAL, LOGL_ERROR,
+			"unable to create socket: %s\n", strerror(errno));
 		return sfd;
+	}
 	if (flags & OSMO_SOCK_F_NONBLOCK) {
 		if (ioctl(sfd, FIONBIO, (unsigned char *)&on) < 0) {
 			LOGP(DLGLOBAL, LOGL_ERROR,
@@ -210,25 +216,35 @@ int osmo_sock_init2(uint16_t family, uint16_t type, uint8_t proto,
 					"cannot setsockopt socket:"
 					" %s:%u: %s\n",
 					local_host, local_port, strerror(errno));
-				break;
+				close(sfd);
+				continue;
 			}
-			if (bind(sfd, rp->ai_addr, rp->ai_addrlen) != -1)
-				break;
-			close(sfd);
+			if (bind(sfd, rp->ai_addr, rp->ai_addrlen) == -1) {
+				LOGP(DLGLOBAL, LOGL_ERROR, "unable to bind socket: %s:%u: %s\n",
+					local_host, local_port, strerror(errno));
+				close(sfd);
+				continue;
+			}
+			break;
 		}
 		freeaddrinfo(result);
 		if (rp == NULL) {
-			LOGP(DLGLOBAL, LOGL_ERROR, "unable to bind socket: %s:%u: %s\n",
-				local_host, local_port, strerror(errno));
+			LOGP(DLGLOBAL, LOGL_ERROR, "no suitable local addr found for: %s:%u\n",
+				local_host, local_port);
 			return -ENODEV;
 		}
 	}
 
+	/* Reached this point, if OSMO_SOCK_F_BIND then sfd is valid (>=0) or it
+	   was already closed and func returned. If OSMO_SOCK_F_BIND is not
+	   set, then sfd = -1 */
+
 	/* figure out remote side of socket */
 	if (flags & OSMO_SOCK_F_CONNECT) {
 		result = addrinfo_helper(family, type, proto, remote_host, remote_port, false);
 		if (!result) {
-			close(sfd);
+			if (sfd >= 0)
+				close(sfd);
 			return -EINVAL;
 		}
 
@@ -246,16 +262,24 @@ int osmo_sock_init2(uint16_t family, uint16_t type, uint8_t proto,
 			}
 
 			rc = connect(sfd, rp->ai_addr, rp->ai_addrlen);
-			if (rc != -1 || (rc == -1 && errno == EINPROGRESS))
-				break;
-
-			close(sfd);
-			sfd = -1;
+			if (rc != 0 && errno != EINPROGRESS) {
+				LOGP(DLGLOBAL, LOGL_ERROR, "unable to connect socket: %s:%u: %s\n",
+					remote_host, remote_port, strerror(errno));
+				/* We want to maintain the bind socket if bind was enabled */
+				if (!(flags & OSMO_SOCK_F_BIND)) {
+					close(sfd);
+					sfd = -1;
+				}
+				continue;
+			}
+			break;
 		}
 		freeaddrinfo(result);
 		if (rp == NULL) {
-			LOGP(DLGLOBAL, LOGL_ERROR, "unable to connect socket: %s:%u: %s\n",
-				remote_host, remote_port, strerror(errno));
+			LOGP(DLGLOBAL, LOGL_ERROR, "no suitable remote addr found for: %s:%u\n",
+				remote_host, remote_port);
+			if (sfd >= 0)
+				close(sfd);
 			return -ENODEV;
 		}
 	}
@@ -316,8 +340,10 @@ int osmo_sock_init(uint16_t family, uint16_t type, uint8_t proto,
 
 		if (flags & OSMO_SOCK_F_CONNECT) {
 			rc = connect(sfd, rp->ai_addr, rp->ai_addrlen);
-			if (rc != -1 || (rc == -1 && errno == EINPROGRESS))
-				break;
+			if (rc != 0 && errno != EINPROGRESS) {
+				close(sfd);
+				continue;
+			}
 		} else {
 			rc = setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR,
 							&on, sizeof(on));
@@ -326,18 +352,24 @@ int osmo_sock_init(uint16_t family, uint16_t type, uint8_t proto,
 					"cannot setsockopt socket:"
 					" %s:%u: %s\n",
 					host, port, strerror(errno));
-				break;
+				close(sfd);
+				continue;
+			}
+			if (bind(sfd, rp->ai_addr, rp->ai_addrlen) == -1) {
+				LOGP(DLGLOBAL, LOGL_ERROR, "unable to bind socket:"
+					"%s:%u: %s\n",
+					host, port, strerror(errno));
+				close(sfd);
+				continue;
 			}
-			if (bind(sfd, rp->ai_addr, rp->ai_addrlen) != -1)
-				break;
 		}
-		close(sfd);
+		break;
 	}
 	freeaddrinfo(result);
 
 	if (rp == NULL) {
-		LOGP(DLGLOBAL, LOGL_ERROR, "unable to connect/bind socket: %s:%u: %s\n",
-			host, port, strerror(errno));
+		LOGP(DLGLOBAL, LOGL_ERROR, "no suitable addr found for: %s:%u\n",
+			host, port);
 		return -ENODEV;
 	}
 
@@ -512,10 +544,32 @@ int osmo_sockaddr_is_local(struct sockaddr *addr, unsigned int addrlen)
 	for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
 		if (!ifa->ifa_addr)
 			continue;
-		if (sockaddr_equal(ifa->ifa_addr, addr, addrlen))
+		if (sockaddr_equal(ifa->ifa_addr, addr, addrlen)) {
+			freeifaddrs(ifaddr);
 			return 1;
+		}
 	}
 
+	freeifaddrs(ifaddr);
+	return 0;
+}
+
+/*! Convert sockaddr to IP address as char string and port as uint16_t.
+ *  \param[out] addr  String buffer to write IP address to, or NULL.
+ *  \param[out] addr_len  Size of \a addr.
+ *  \param[out] port  Pointer to uint16_t to write the port number to, or NULL.
+ *  \param[in] sa  Sockaddr to convert.
+ *  \returns the required string buffer size, like osmo_strlcpy(), or 0 if \a addr is NULL.
+ */
+unsigned int osmo_sockaddr_to_str_and_uint(char *addr, unsigned int addr_len, uint16_t *port,
+					   const struct sockaddr *sa)
+{
+	const struct sockaddr_in *sin = (const struct sockaddr_in *)sa;
+
+	if (port)
+		*port = ntohs(sin->sin_port);
+	if (addr)
+		return osmo_strlcpy(addr, inet_ntoa(sin->sin_addr), addr_len);
 	return 0;
 }
 
@@ -825,12 +879,13 @@ int osmo_sock_local_ip(char *local_ip, const char *remote_ip)
 	memset(&local_addr, 0, sizeof(local_addr));
 	local_addr_len = sizeof(local_addr);
 	rc = getsockname(sfd, (struct sockaddr *)&local_addr, &local_addr_len);
+	close(sfd);
 	if (rc < 0)
 		return -EINVAL;
 	if (local_addr.sin_family == AF_INET)
-		strncpy(local_ip, inet_ntoa(local_addr.sin_addr), INET_ADDRSTRLEN);
+		inet_ntop(AF_INET, &local_addr.sin_addr, local_ip, INET_ADDRSTRLEN);
 	else if (local_addr.sin_family == AF_INET6)
-		strncpy(local_ip, inet_ntoa(local_addr.sin_addr), INET6_ADDRSTRLEN);
+		inet_ntop(AF_INET6, &local_addr.sin_addr, local_ip, INET6_ADDRSTRLEN);
 	else
 		return -EINVAL;
 
