@@ -3,6 +3,12 @@
 
    Command interpreter routine for virtual terminal [aka TeletYpe]
    Copyright (C) 1997, 98, 99 Kunihiro Ishiguro
+   Copyright (C) 2010-2011 Holger Hans Peter Freyther <zecke@selfish.org>
+   Copyright (C) 2012 Sylvain Munaut <tnt@246tNt.com>
+   Copyright (C) 2013,2015 Harald Welte <laforge@gnumonks.org>
+   Copyright (C) 2013,2017 sysmocom - s.f.m.c. GmbH
+
+   SPDX-License-Identifier: GPL-2.0+
 
 This file is part of GNU Zebra.
 
@@ -62,21 +68,25 @@ struct host host;
 struct cmd_node auth_node = {
 	AUTH_NODE,
 	"Password: ",
+	.name = "auth",
 };
 
 struct cmd_node view_node = {
 	VIEW_NODE,
 	"%s> ",
+	.name = "view",
 };
 
 struct cmd_node auth_enable_node = {
 	AUTH_ENABLE_NODE,
 	"Password: ",
+	.name = "auth-enable",
 };
 
 struct cmd_node enable_node = {
 	ENABLE_NODE,
 	"%s# ",
+	.name = "enable",
 };
 
 struct cmd_node config_node = {
@@ -504,7 +514,7 @@ char *osmo_asciidoc_escape(const char *inp)
 		}
 	}
 
-	out = talloc_size(NULL, len + 1);
+	out = talloc_size(tall_vty_cmd_ctx, len + 1);
 	if (!out)
 		return NULL;
 
@@ -564,7 +574,7 @@ static char *xml_escape(const char *inp)
 		}
 	}
 
-	out = talloc_size(NULL, len + 1);
+	out = talloc_size(tall_vty_cmd_ctx, len + 1);
 	if (!out)
 		return NULL;
 
@@ -639,6 +649,8 @@ static int vty_dump_element(struct cmd_element *cmd, struct vty *vty)
 	return 0;
 }
 
+static bool vty_command_is_common(struct cmd_element *cmd);
+
 /*
  * Dump all nodes and commands associated with a given node as XML to the VTY.
  */
@@ -649,11 +661,37 @@ static int vty_dump_nodes(struct vty *vty)
 
 	vty_out(vty, "<vtydoc xmlns='urn:osmocom:xml:libosmocore:vty:doc:1.0'>%s", VTY_NEWLINE);
 
+	/* Only once, list all common node commands. Use the CONFIG node to find common node commands. */
+	vty_out(vty, "  <node id='_common_cmds_'>%s", VTY_NEWLINE);
+	vty_out(vty, "    <name>Common Commands</name>%s", VTY_NEWLINE);
+	vty_out(vty, "    <description>These commands are available on all VTY nodes. They are listed"
+		" here only once, to unclutter the VTY reference.</description>%s", VTY_NEWLINE);
 	for (i = 0; i < vector_active(cmdvec); ++i) {
 		struct cmd_node *cnode;
 		cnode = vector_slot(cmdvec, i);
 		if (!cnode)
 			continue;
+		if (cnode->node != CONFIG_NODE)
+			continue;
+
+		for (j = 0; j < vector_active(cnode->cmd_vector); ++j) {
+			struct cmd_element *elem;
+			elem = vector_slot(cnode->cmd_vector, j);
+			if (!vty_command_is_common(elem))
+				continue;
+			if (!elem->attr & CMD_ATTR_DEPRECATED)
+				vty_dump_element(elem, vty);
+		}
+	}
+	vty_out(vty, "  </node>%s", VTY_NEWLINE);
+
+	for (i = 0; i < vector_active(cmdvec); ++i) {
+		struct cmd_node *cnode;
+		cnode = vector_slot(cmdvec, i);
+		if (!cnode)
+			continue;
+		if (vector_active(cnode->cmd_vector) < 1)
+			continue;
 
 		/* De-dup node IDs: how many times has this same name been used before? Count the first
 		 * occurence as _1 and omit that first suffix, so that the first occurence is called
@@ -672,11 +710,15 @@ static int vty_dump_nodes(struct vty *vty)
 		if (same_name_count > 1 || !*cnode->name)
 			vty_out(vty, "_%d", same_name_count);
 		vty_out(vty, "'>%s", VTY_NEWLINE);
+		vty_out(vty, "    <name>%s</name>%s", cnode->name, VTY_NEWLINE);
 
 		for (j = 0; j < vector_active(cnode->cmd_vector); ++j) {
 			struct cmd_element *elem;
 			elem = vector_slot(cnode->cmd_vector, j);
-			vty_dump_element(elem, vty);
+			if (vty_command_is_common(elem))
+				continue;
+			if (!elem->attr & CMD_ATTR_DEPRECATED)
+				vty_dump_element(elem, vty);
 		}
 
 		vty_out(vty, "  </node>%s", VTY_NEWLINE);
@@ -1262,25 +1304,17 @@ static int cmd_range_match(const char *range, const char *str)
 }
 
 /* helper to retrieve the 'real' argument string from an optional argument */
-static char *
-cmd_deopt(const char *str)
+static char *cmd_deopt(void *ctx, const char *str)
 {
 	/* we've got "[blah]". We want to strip off the []s and redo the
 	 * match check for "blah"
 	 */
 	size_t len = strlen(str);
-	char *tmp;
 
 	if (len < 3)
 		return NULL;
 
-	/* tmp will hold a string of len-2 chars, so 'len' size is fine */
-	tmp = talloc_size(NULL, len);
-
-	memcpy(tmp, (str + 1), len - 2);
-	tmp[len - 2] = '\0';
-
-	return tmp;
+	return talloc_strndup(ctx, str + 1, len - 2);
 }
 
 static enum match_type
@@ -1291,7 +1325,7 @@ cmd_match(const char *str, const char *command,
 	if (recur && CMD_OPTION(str))
 	{
 		enum match_type ret;
-		char *tmp = cmd_deopt(str);
+		char *tmp = cmd_deopt(tall_vty_cmd_ctx, str);
 
 		/* this would be a bug in a command, however handle it gracefully
 		 * as it we only discover it if a user tries to run it
@@ -1440,6 +1474,7 @@ cmd_filter(char *command, vector v, unsigned int index, enum match_type level)
 static int
 is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 {
+	int ret = 0;
 	unsigned int i;
 	unsigned int j;
 	struct cmd_element *cmd_element;
@@ -1447,6 +1482,16 @@ is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 	vector descvec;
 	struct desc *desc;
 
+	/* In this loop, when a match is found, 'matched' points to it. If on a later iteration, an
+	 * identical match is found, the command is ambiguous. The trickiness is that a string may be
+	 * enclosed in '[str]' square brackets, which get removed by a talloc_strndup(), via cmd_deopt().
+	 * Such a string is usually needed for one loop iteration, except when 'matched' points to it. In
+	 * that case, the string must remain allocated until this function exits or another match comes
+	 * around. This is sufficiently confusing to justify a separate talloc tree to store all of the
+	 * odd allocations, and to free them all at the end. We are not expecting too many optional args
+	 * or ambiguities to cause a noticeable memory footprint from keeping all allocations. */
+	void *cmd_deopt_ctx = NULL;
+
 	for (i = 0; i < vector_active(v); i++)
 		if ((cmd_element = vector_slot(v, i)) != NULL) {
 			int match = 0;
@@ -1458,9 +1503,15 @@ is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 					enum match_type ret;
 					const char *str = desc->cmd;
 
-					if (CMD_OPTION(str))
-						if ((str = cmd_deopt(str)) == NULL)
+					if (CMD_OPTION(str)) {
+						if (!cmd_deopt_ctx)
+							cmd_deopt_ctx =
+								talloc_named_const(tall_vty_cmd_ctx, 0,
+										   __func__);
+						str = cmd_deopt(cmd_deopt_ctx, str);
+						if (str == NULL)
 							continue;
+					}
 
 					switch (type) {
 					case exact_match:
@@ -1474,9 +1525,10 @@ is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 						{
 							if (matched
 							    && strcmp(matched,
-								      str) != 0)
-								return 1;	/* There is ambiguous match. */
-							else
+								      str) != 0) {
+								ret = 1; /* There is ambiguous match. */
+								goto free_and_return;
+							} else
 								matched = str;
 							match++;
 						}
@@ -1486,9 +1538,10 @@ is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 						    (str, command)) {
 							if (matched
 							    && strcmp(matched,
-								      str) != 0)
-								return 1;
-							else
+								      str) != 0) {
+								ret = 1;
+								goto free_and_return;
+							} else
 								matched = str;
 							match++;
 						}
@@ -1502,8 +1555,10 @@ is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 						if ((ret =
 						     cmd_ipv6_prefix_match
 						     (command)) != no_match) {
-							if (ret == partly_match)
-								return 2;	/* There is incomplete match. */
+							if (ret == partly_match) {
+								ret = 2;	/* There is incomplete match. */
+								goto free_and_return;
+							}
 
 							match++;
 						}
@@ -1517,8 +1572,10 @@ is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 						if ((ret =
 						     cmd_ipv4_prefix_match
 						     (command)) != no_match) {
-							if (ret == partly_match)
-								return 2;	/* There is incomplete match. */
+							if (ret == partly_match) {
+								ret = 2;	/* There is incomplete match. */
+								goto free_and_return;
+							}
 
 							match++;
 						}
@@ -1531,14 +1588,15 @@ is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 					default:
 						break;
 					}
-
-					if (CMD_OPTION(desc->cmd))
-						talloc_free((void*)str);
 				}
 			if (!match)
 				vector_slot(v, i) = NULL;
 		}
-	return 0;
+
+free_and_return:
+	if (cmd_deopt_ctx)
+		talloc_free(cmd_deopt_ctx);
+	return ret;
 }
 
 /* If src matches dst return dst string, otherwise return NULL */
@@ -1950,7 +2008,7 @@ static char **cmd_complete_command_real(vector vline, struct vty *vty,
 
 		/* In case of 'command \t' pattern.  Do you need '?' command at
 		   the end of the line. */
-		if (vector_slot(vline, index) == '\0')
+		if (vector_slot(vline, index) == NULL)
 			*status = CMD_ERR_NOTHING_TODO;
 		else
 			*status = CMD_ERR_NO_MATCH;
@@ -2688,16 +2746,26 @@ static int write_config_file(const char *config_file, char **outpath)
 
 	*outpath = NULL;
 
+	/* The string composition code here would be a case for talloc_asprintf(), but the pseudotalloc.c
+	 * talloc_asprintf() implementation would truncate a too-long path with "[...]", so doing it
+	 * manually instead. */
+
 	/* Check and see if we are operating under vtysh configuration */
 	config_file_sav =
 	    _talloc_zero(tall_vty_cmd_ctx,
 			 strlen(config_file) + strlen(CONF_BACKUP_EXT) + 1,
 			 "config_file_sav");
+	if (!config_file_sav)
+		return -1;
 	strcpy(config_file_sav, config_file);
 	strcat(config_file_sav, CONF_BACKUP_EXT);
 
 	config_file_tmp = _talloc_zero(tall_vty_cmd_ctx, strlen(config_file) + 8,
-					"config_file_tmp");
+				       "config_file_tmp");
+	if (!config_file_tmp) {
+		talloc_free(config_file_sav);
+		return -1;
+	}
 	sprintf(config_file_tmp, "%s.XXXXXX", config_file);
 
 	/* Open file to configuration write. */
@@ -3642,6 +3710,24 @@ static void install_basic_node_commands(int node)
 	}
 }
 
+/*! Return true if a node is installed by install_basic_node_commands(), so
+ * that we can avoid repeating them for each and every node during 'show
+ * running-config' */
+static bool vty_command_is_common(struct cmd_element *cmd)
+{
+	if (cmd == &config_help_cmd
+	    || cmd == &config_list_cmd
+	    || cmd == &config_write_terminal_cmd
+	    || cmd == &config_write_file_cmd
+	    || cmd == &config_write_memory_cmd
+	    || cmd == &config_write_cmd
+	    || cmd == &show_running_config_cmd
+	    || cmd == &config_exit_cmd
+	    || cmd == &config_end_cmd)
+		return true;
+	return false;
+}
+
 /**
  * Write the current running config to a given file
  * \param[in] vty the vty of the code
