@@ -1,9 +1,11 @@
 /*
  * (C) 2009-2010 by Harald Welte <laforge@gnumonks.org>
  * (C) 2009-2014 by Holger Hans Peter Freyther
- * (C) 2015      by Sysmocom s.f.m.c. GmbH
+ * (C) 2015      by sysmocom - s.f.m.c. GmbH
  * All Rights Reserved
  *
+ * SPDX-License-Identifier: GPL-2.0+
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -391,6 +393,13 @@ static int asciidoc_handle_counter(struct osmo_counter *counter, void *sctx_)
 
 static void asciidoc_counter_generate(struct vty *vty)
 {
+	if (osmo_counters_count() == 0)
+	{
+		vty_out(vty, "// there are no ungrouped osmo_counters%s",
+			VTY_NEWLINE);
+		return;
+	}
+
 	vty_out(vty, "// ungrouped osmo_counters%s", VTY_NEWLINE);
 	vty_out(vty, ".ungrouped osmo counters%s", VTY_NEWLINE);
 	vty_out(vty, "[options=\"header\"]%s", VTY_NEWLINE);
@@ -507,17 +516,36 @@ DEFUN(show_stats_asciidoc_table,
 		host.app_info->name ? host.app_info->name : "", VTY_NEWLINE, VTY_NEWLINE);
 	/* 2x VTY_NEWLINE are intentional otherwise it would interpret the first table header
 	 * as usual text*/
+	vty_out(vty, "=== Rate Counters%s%s", VTY_NEWLINE, VTY_NEWLINE);
 	vty_out(vty, "// generating tables for rate_ctr_group%s", VTY_NEWLINE);
 	rate_ctr_for_each_group(asciidoc_rate_ctr_group_handler, vty);
 
+	vty_out(vty, "== Osmo Stat Items%s%s", VTY_NEWLINE, VTY_NEWLINE);
 	vty_out(vty, "// generating tables for osmo_stat_items%s", VTY_NEWLINE);
 	osmo_stat_item_for_each_group(asciidoc_osmo_stat_item_group_handler, vty);
 
+	vty_out(vty, "== Osmo Counters%s%s", VTY_NEWLINE, VTY_NEWLINE);
 	vty_out(vty, "// generating tables for osmo_counters%s", VTY_NEWLINE);
 	asciidoc_counter_generate(vty);
 	return CMD_SUCCESS;
 }
 
+static int rate_ctr_group_handler(struct rate_ctr_group *ctrg, void *sctx_)
+{
+	struct vty *vty = sctx_;
+	vty_out_rate_ctr_group_fmt(vty, "%25n: %10c (%S/s %M/m %H/h %D/d) %d", ctrg);
+	return 0;
+}
+
+DEFUN(show_rate_counters,
+      show_rate_counters_cmd,
+      "show rate-counters",
+      SHOW_STR "Show all rate counters\n")
+{
+	rate_ctr_for_each_group(rate_ctr_group_handler, vty);
+	return CMD_SUCCESS;
+}
+
 static int config_write_stats_reporter(struct vty *vty, struct osmo_stats_reporter *srep)
 {
 	if (srep == NULL)
@@ -611,4 +639,5 @@ void osmo_stats_vty_add_cmds()
 	install_element(CFG_STATS_NODE, &cfg_stats_reporter_disable_cmd);
 
 	install_element_ve(&show_stats_asciidoc_table_cmd);
+	install_element_ve(&show_rate_counters_cmd);
 }
