@@ -22,7 +22,7 @@
  *
  * \section sec_copyright Copyright and License
  * Copyright © 1997-2007 - Kuninhiro Ishiguro\n
- * Copyright © 2008-2011 - Harald Welte, Holger Freyther and contributors\n
+ * Copyright © 2008-2012 - Harald Welte, Holger Freyther and contributors\n
  * All rights reserved. \n\n
  * The source code of libosmovty is licensed under the terms of the GNU
  * General Public License as published by the Free Software Foundation;
@@ -46,6 +46,7 @@
  * <http://sysmocom.de/>
  */
 
+/* SPDX-License-Identifier: GPL-2.0+ */
 
 #include <stdio.h>
 #include <stdarg.h>
@@ -361,7 +362,7 @@ void vty_hello(struct vty *vty)
 	if (host.app_info->name)
 		app_name = host.app_info->name;
 
-	vty_out(vty, "Welcome to the %s control interface%s%s",
+	vty_out(vty, "Welcome to the %s VTY interface%s%s",
 		app_name, VTY_NEWLINE, VTY_NEWLINE);
 
 	if (host.app_info->copyright)
@@ -1505,7 +1506,7 @@ vty_create (int vty_sock, void *priv)
 {
   struct vty *vty;
 
-	struct termios t;
+	struct termios t = {};
 
 	tcgetattr(vty_sock, &t);
 	cfmakeraw(&t);
