@@ -1,6 +1,8 @@
 /* test routines for NS connection handling
- * (C) 2013 by sysmocom s.f.m.c. GmbH
+ * (C) 2013 by sysmocom - s.f.m.c. GmbH
  * Author: Jacob Erlbeck <jerlbeck@sysmocom.de>
+ *
+ * SPDX-License-Identifier: GPL-2.0+
  */
 
 #undef _GNU_SOURCE
@@ -899,7 +901,8 @@ static struct log_info info = {};
 
 int main(int argc, char **argv)
 {
-	osmo_init_logging(&info);
+	void *ctx = talloc_named_const(NULL, 0, "gprs_ns_test");
+	osmo_init_logging2(ctx, &info);
 	log_set_use_color(osmo_stderr_target, 0);
 	log_set_print_filename(osmo_stderr_target, 0);
 	osmo_signal_register_handler(SS_L_NS, &test_signal, NULL);
